#!/usr/bin/env python
# coding=utf8
#
# Copyright (C) 2010  Edmundo Álvarez Jiménez
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Edmundo Álvarez Jiménez <e.alvarezj@gmail.com>

import os
import time
from Squealer import defs
from Squealer.FileLocker import FileLocker
from Squealer.settings import SettingsReader
from xml.dom import Node
from xml.dom import minidom
from datetime import datetime

class Event:
    """Represents a single event tracked by TempusFugit"""
    
    def __init__ (self, itime, etime, app, title):
        self.init_time = datetime.fromtimestamp (int(itime))
        self.end_time = datetime.fromtimestamp (int(etime))
        self.app = app
        self.title = title
    
    def __str__ (self):
        return ("Event %s: %s") % (self.init_time, self.app)
    
    def get_initial_time (self):
        return str (self.init_time)
    
    def get_length (self):
        return str (self.end_time - self.init_time)
    
    def get_app (self):
        return self.app
    
    def get_title (self):
        return self.title

class EventsParser:
    """Parse events' file"""
    
    def __init__ (self):
        self.locker = FileLocker ()
        
        self.settings = SettingsReader ()
        events_path = self.settings.get_events_file ()
        
        self.events_file = os.path.join (os.getenv ("HOME"), ".detras", \
                events_path)
    
    
    def parse (self):
        """Convert an XML file containing events in an Events' list"""
        self.locker.lock (self.events_file)
        try:
            xml_doc = minidom.parse (self.events_file)
            self.locker.unlock (self.events_file)
            
            events_nodes = xml_doc.getElementsByTagName ("event")
            events = []
            
            for i in events_nodes:
                if (i.hasAttribute ("type")) \
                        and (i.getAttribute ("type") == "WindowChange"):
                    i.normalize ()
                    
                    itime = i.getAttribute ("time")
                    
                    next_event = i.nextSibling
                    
                    while (next_event is not None) \
                            and (next_event.nodeType != Node.ELEMENT_NODE):
                        next_event = next_event.nextSibling
                    
                    if next_event is None:
                        etime = int (time.time ())
                    else:
                        etime = next_event.getAttribute ("time")
                    
                    app_child = i.getElementsByTagName ("app") \
                            [0].firstChild
                    
                    if app_child is not None:
                        app = app_child.data
                    
                    title_child = i.getElementsByTagName ("title") \
                            [0].firstChild
                    
                    if title_child is not None:
                        title = title_child.data
                    
                    events.append (Event (itime, etime, app, title))
            
            xml_doc.unlink ()
        
            return events
        except IOError:
            return ""
