#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Authors: Carlos Garcia Campos <carlosgc@gsyc.escet.urjc.es>

from PyDataAccess.ConnectionFactory import ConnectionFactory as cf

# Table_Field
(
    EVENTS_ID,
    EVENTS_CLIENT,
    EVENTS_TIME,
    EVENTS_TYPE,
    EVENTS_EVENT,
    EVENTS_DURATION
) = range (6)


class Activity:

    def __init__ (self, db, id, timestamp, duration):
        self.db = db

        self.id = id
        self.timestamp = timestamp
        self.app = None
        self.title = None
        if duration is None:
            self.duration = -1
        else:
            self.duration = duration

    def __str__ (self):
        return "%s::%s - %d" % (self.get_application (), self.get_title (), self.duration)

    def __repr__ (self):
        return self.__str__ ()

    def get_application (self):
        if self.app is not None:
            return self.app

        rs = self.db.execute ("select application from EventWindow where id = ?", (self.id,))
        self.app = rs[0][0]
        return self.app

    def get_title (self):
        if self.title is not None:
            return self.title

        rs = self.db.execute ("select title from EventWindow where id = ?", (self.id,))
        self.title = rs[0][0]
        return self.title

    def get_duration (self):
        return self.duration

    def get_date (self):
        from time import strftime,gmtime
        return strftime ("%d/%m/%Y %H:%M", gmtime (int (self.timestamp)))


class Reports:

    def __init__ (self, dbfile):
        self.db = cf.create_connector ("SQLite")
        self.db.connect (db=dbfile)

        self.event_types = []
        rs = self.db.execute ("select * from EventTypes")
        for type in rs:
            self.event_types.append (type[1])

    def __get_event_type (self, type):
        return self.event_types[int (type) - 1]

    def __get_event_type_id (self, type):
        try:
            retval = self.event_types.index (type)
        except:
            return -1

        return retval + 1

    def __get_activities (self, rs):
        retval = {}

        for event in rs:
            event_type = self.__get_event_type (event[EVENTS_TYPE])

            if event_type != 'Window':
                continue

            client = event[EVENTS_CLIENT]
            if not retval.has_key (client):
                retval[client] = []
            
            retval[client].append (Activity (self.db, event[EVENTS_EVENT], event[EVENTS_TIME], event[EVENTS_DURATION]))
           
        return retval

    def __get_event_id (self, app, title):
        if app is not None and title is not None:
            command = "select id from EventWindow where application like ? and title like ?"
            args = (app, title)
        elif app is not None:
            command = "select id from EventWindow where application like ?"
            args = (app,)
        elif title is not None:
            command = "select id from EventWindow where title like ?"
            args = (title,)
        else:
            return []

        rs = self.db.execute (command, args)
        if len (rs) <= 0:
            return []
        
        ids = []
        for item in rs:
            ids.append (item[0])

        return ids

    def __build_condition (self, id, app, title, timeval):
        command = ""
        args = ()

        if id is not None:
            command += "client_id = ?"
            args += (id,)

        if timeval is not None:
            if id is not None:
                command += " and "
            command += " (timestamp >= ? and timestamp <= ?) "
            args += timeval

        events = self.__get_event_id (app, title)
        if len (events) <= 0:
            return ("", ())

        if id is not None or timeval is not None:
            command += " and "
        
        command += " ("

        i = 0
        for event in events:
            if i > 0:
                command += " or event = ?"
            else:
                command += "event = ?"

            args += (event,)
            i +=1

        command += ")"

        return (command, args)

    def get_activities (self, id = None, app = None, title = None, timeval = None):
        command = "select * from Events "
        args = ()
        (cond, cond_args) = self.__build_condition (id, app, title, timeval)
        if cond != "":
            command += " where "
            command += cond
            args += cond_args

        command += " order by %d,%d" % (EVENTS_CLIENT, EVENTS_TIME)

        rs = self.db.execute (command, args)
        return self.__get_activities (rs)

    def get_applications (self):
        retval = []

        command = "select distinct (application) from EventWindow"
        rs = self.db.execute (command)
       
        for item in rs:
            retval.append (item[0])

        return retval

    def get_cumulative_duration (self, id = None, app = None, title = None, timeval = None):
        command = "select sum(duration) from Events where event_type = ? "
        args = (self.__get_event_type_id ('Window'),)
        (cond, cond_args) = self.__build_condition (id, app, title, timeval)
        if (app is not None or title is not None) and cond == "":
            return 0

        if cond != "":
            command += " and "
            command += cond
            args += cond_args

        retval = self.db.execute (command, args)[0][0]
        if retval is None:
            return 0

        return retval


if __name__ == '__main__':

    import sys

    def print_activities (acts, nmax = sys.maxint):
        if acts is None:
            return

        for client in acts:
            print "Client %s" % (client)
            print
            for i in range (min (nmax, len (acts[client]))):
                activity = acts[client][i]
                print "%s %s  \"%s\" %d" % (activity.get_date (), activity.get_application (), activity.get_title (), activity.get_duration ())

    rep = Reports ("./data/detras.db")

    total = rep.get_cumulative_duration ()

    apps = rep.get_applications ()
    for app in apps:
        parc = rep.get_cumulative_duration (app = app)
        print "%s: %d (%.2f%%)" % (app, parc, (parc * 100) / float (total))
        
    acts = rep.get_activities ()
    print_activities (acts, 50)

    acts = rep.get_activities (app = "Gnome-Terminal%")
    print_activities (acts, 10)
    
    acts = rep.get_activities (app = "Evolution%")
    print_activities (acts, 10)
    
    

