#!/usr/bin/env python
# coding=utf8
#
# Copyright (C) 2008-2009 Adam Olsen
# Copyright (C) 2010 Edmundo Álvarez Jiménez
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Settings
#
# stores settings


from ConfigParser import RawConfigParser, NoSectionError, NoOptionError
import logging
import os
import sys

import defs
from FileLocker import FileLocker

logger = logging.getLogger(__name__)

TYPE_MAPPING = {
        'I': int,
        'S': str,
        'F': float,
        'B': bool,
        'L': list,
        'D': dict,
        'U': unicode
        }

class SettingsManager(RawConfigParser):
    """
        Manages settings. Code taken from Exaile and adapted to fit DeTraS.
    """
    
    READ_MODE = 'r'
    EDIT_MODE = 'e'
    
    settings = None
    __version__ = 1
    def __init__(self, loc, defaultloc=None, \
            mode=READ_MODE):
        """
            Sets up the SettingsManager. Expects a loc to a file
            where settings will be stored.

            If loc is None, these settings will never be stored nor read from a
            file

            defaultloc is a loc to initialize settings from
            
            mode indicates if SettingsMgr is going to read settings
            (READ_MODE) or modify them (EDIT_MODE)
        """
        RawConfigParser.__init__(self)
        self.loc = loc
        self._saving = False
        self._dirty = False
        self.mode = mode
        
        self.locker = FileLocker()
        
        if defaultloc is not None:
            try:
                self.read(defaultloc)
            except:
                pass

        if self.loc is not None:
            if (not os.path.exists(self.loc)) \
                    and (self.mode is SettingsManager.EDIT_MODE):
                try:
                    dirname = os.path.dirname(self.loc)
                    
                    try:
                        os.makedirs(dirname)
                    except:
                        print "'%s' already exists" % dirname
                    
                    fd = open(self.loc, 'w')
                    fd.close()
                except:
                    eType, eValue, eTraceback = sys.exc_info ()
                    print "Exception found: %s" % eValue
            
            if os.path.exists(self.loc):
                self.locker.lock(self.loc)
                
                self.read(self.loc)
                
                if self.mode is SettingsManager.READ_MODE:
                    self.locker.unlock(self.loc)
            
        if not self.get_option('settings/version', 0):
            self.set_option('settings/version', self.__version__)

    def _timeout_save(self):
        self.save()
        return True
    
    def finished(self):
        if self.mode is SettingsManager.EDIT_MODE:
            self.locker.unlock(self.loc)

    def copy_settings(self, settings):
        """
            Copies one all of the settings contained in this instance to
            another
        """
        for section in self.sections():
            for (key, value) in self.items(section):
                settings._set_direct('%s/%s' % (section, key), value)

    def clone(self):
        """
            Creates a copy of this settings object
        """
        settings = SettingsManager(None)
        self.copy_settings(settings)
        return settings

    def set_option(self, option, value):
        """
            Set an option (in section/key syntax) to the specified value.
        """
        value = self._val_to_str(value)
        splitvals = option.split('/')
        section, key = "/".join(splitvals[:-1]), splitvals[-1]
        try:
            self.set(section, key, value)
        except NoSectionError:
            self.add_section(section)
            self.set(section, key, value)
        self._dirty = True

    def get_option(self, option, default=None):
        """
            Get the value of an option (in section/key syntax), returning
            default if the key does not exist yet.
        """
        splitvals = option.split('/')
        section, key = "/".join(splitvals[:-1]), splitvals[-1]
        try:
            value = self.get(section, key)
            value = self._str_to_val(value)
        except NoSectionError:
            value = default
        except NoOptionError:
            value = default
        return value

    def _set_direct(self, option, value):
        """
            Sets the option directly to the value, only for use in copying
            settings.
        """
        splitvals = option.split('/')
        section, key = "/".join(splitvals[:-1]), splitvals[-1]
        try:
            self.set(section, key, value)
        except NoSectionError:
            self.add_section(section)
            self.set(section, key, value)

    def _val_to_str(self, value):
        """
            Turns a value of some type into a string so it
            can be a configuration value.
        """
        for k, v in TYPE_MAPPING.iteritems():
            if v == type(value):
                if v == list:
                    return k + ": " + repr(value)
                else:
                    return k + ": " + str(value)
        raise ValueError("We don't know how to store that kind of setting: ",
            type(value))

    def _str_to_val(self, value):
        """
            Convert setting strings back to normal values.
        """
        try:
            kind, value = value.split(': ', 1)
        except ValueError:
            return ''

        # Lists are special case
        if kind == 'L':
            return eval(value)

        # So are dictionaries
        if kind == 'D':
            return eval(value)

        if kind in TYPE_MAPPING.keys():
            if kind == 'B':
                if value != 'True':
                    return False
            value = TYPE_MAPPING[kind](value)
            return value
        else:
            raise ValueError("An Unknown type of setting was found!")

    def save(self):
        """
            Save the settings to disk
        """
        if self.loc is None:
            logger.debug("Save requested but not saving settings, loc is None")
            return
        if self._saving or not self._dirty:
            return

        self._saving = True
        
        logger.debug("Saving settings...")
        with open(self.loc, 'w') as f:
            self.write(f)
            f.close()
        
        self._saving = False
        self._dirty = False
        

class SettingsReader:
    """
        Provides methods to read settings using SettingsManager.
        It has the advantage of being less settings depending.
    """
    
    def __init__ (self):
        import GUID
        
        self.settings = SettingsManager (defs.USER_CFG_FILE, \
                defs.DEFAULT_CFG_FILE, SettingsManager.READ_MODE)
        
        client_id_file = os.path.join (defs.USER_DATA_DIR, "id")
        if not os.path.exists (client_id_file):
            if not os.path.exists (defs.USER_DATA_DIR):
                os.makedirs (defs.USER_DATA_DIR)
            
            uid = GUID.generate ()
            
            f = open (client_id_file, "w")
            f.write (uid)
            f.close ()
            
            self.client_id = uid
        else:
            f = open (client_id_file, "r")
            self.client_id = f.read ()
            f.close ()
        
    def get_client_id (self):
        return self.client_id
    
    def get_client_name (self):
        return self.settings.get_option ("client/name", "")
    
    def get_client_projects (self):
        return self.settings.get_option ("client/projects", [])
    
    def get_events_file (self):
        return os.path.join (defs.USER_DATA_DIR, \
                self.settings.get_option ("general/events_file", \
                "tempusfugit.xml"))
    
    def get_start_init (self):
        return self.settings.get_option ("general/start_init", False)
    
    def get_stop_exit (self):
        return self.settings.get_option ("general/stop_exit", True)
    
    def get_server_url (self):
        server_addr = self.settings.get_option ("server/address", \
                "https://localhost")
        server_port = self.settings.get_option ("server/port", 8080)
        service_name = self.settings.get_option ("server/service_name", \
                "DeTraS")
        
        return ("%s:%d/%s") % (server_addr, server_port, service_name)
    
    def get_send_client_name (self):
        return self.settings.get_option ("privacy/send_client_name", False)
    
    def get_filter_type (self):
        return self.settings.get_option ("privacy/filter_type", 0)
    
    def get_re_filter_patterns (self):
        return self.settings.get_option ("privacy/re_filter_patterns", [])


class SettingsWritter:
    """
        Read or write settings using SettingsManager.
        It's less settings dependent.
        There must be only one SettingsWritter managing a file.
    """
    
    def __init__ (self):
        import GUID
        
        self.settings = SettingsManager (defs.USER_CFG_FILE, \
                defs.DEFAULT_CFG_FILE, SettingsManager.EDIT_MODE)
        
        client_id_file = os.path.join (defs.USER_DATA_DIR, "id")
        if not os.path.exists (client_id_file):
            uid = GUID.generate ()
            
            f = open (client_id_file, "w")
            f.write (uid)
            f.close ()
            
            self.client_id = uid
        else:
            f = open (client_id_file, "r")
            self.client_id = f.read ()
            f.close ()
        
    def __set_option (self, option, value):
        self.settings.set_option (option, value)
        self.settings.save ()
    
    def get_client_id (self):
        return self.client_id
    
    def get_client_name (self):
        return self.settings.get_option ("client/name", "")
    
    def set_client_name (self, value):
        self.__set_option ("client/name", value)
    
    def get_client_projects (self):
        return self.settings.get_option ("client/projects", [])
    
    def set_client_projects (self, value):
        self.__set_option ("client/projects", value)
    
    def get_events_file (self):
        return os.path.join (defs.USER_DATA_DIR, \
                self.settings.get_option ("general/events_file", \
                "tempusfugit.xml"))
    
    def get_start_init (self):
        return self.settings.get_option ("general/start_init", False)
    
    def set_start_init (self, value):
        self.__set_option ("general/start_init", value)
    
    def get_stop_exit (self):
        return self.settings.get_option ("general/stop_exit", True)
    
    def set_stop_exit (self, value):
        self.__set_option ("general/stop_exit", value)
    
    def get_server_url (self):
        server_addr = self.settings.get_option ("server/address", \
                "https://localhost")
        server_port = self.settings.get_option ("server/port", 8080)
        service_name = self.settings.get_option ("server/service_name", \
                "DeTraS")
        return ("%s:%d/%s") % (server_addr, server_port, service_name)
    
    def get_server_address (self):
        return self.settings.get_option ("server/address", "https://localhost")
    
    def set_server_address (self, value):
        self.__set_option ("server/address", value)
    
    def get_server_port (self):
        return self.settings.get_option ("server/port", 8080)
    
    def set_server_port (self, value):
        self.__set_option ("server/port", value)
    
    def get_service_name (self):
        return self.settings.get_option ("server/service_name", "DeTraS")
    
    def set_service_name (self, value):
        self.__set_option ("server/service_name", value)
    
    def get_send_client_name (self):
        return self.settings.get_option ("privacy/send_client_name", False)
    
    def set_send_client_name (self, value):
        self.__set_option ("privacy/send_client_name", value)
    
    def get_filter_type (self):
        return self.settings.get_option ("privacy/filter_type", 0)
    
    def set_filter_type (self, value):
        self.__set_option ("privacy/filter_type", value)
    
    def get_re_filter_patterns (self):
        return self.settings.get_option ("privacy/re_filter_patterns", [])
    
    def set_re_filter_patterns (self, value):
        self.__set_option ("privacy/re_filter_patterns", value)
    
    def finished (self):
        self.settings.finished ()

