/*
 *  Copyright (C) 2006 Carlos Garcia Campos
 *  Copyright (C) 2010 Edmundo Álvarez Jiménez
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Authors: Carlos Garcia Campos <carlosgc@gsyc.escet.urjc.es>
 *           Edmundo Álvarez Jiménez <e.alvarezj@gmail.com>
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gtk/gtktooltips.h>
#include <gtk/gtkimage.h>
#include <gtk/gtkstock.h>
#include <gtk/gtkaboutdialog.h>
#include <gtk/gtkicontheme.h>
#include <gtk/gtkaccessible.h>
#include <panel-applet.h>

#include "tf-daemon.h"

#define TF_TYPE_APPLET     (tf_applet_get_type ())
#define TF_APPLET(obj)     (G_TYPE_CHECK_INSTANCE_CAST ((obj), TF_TYPE_APPLET, TfApplet))
#define TF_IS_APPLET(obj)  (G_TYPE_CHECK_INSTANCE_TYPE ((obj), TF_TYPE_APPLET))


typedef struct _TfApplet      TfApplet;
typedef struct _TfAppletClass TfAppletClass;

struct _TfApplet {
	PanelApplet      base;

	GtkWidget       *image;
	GtkTooltips     *tips;

	guint            source_id;
	TfDaemonStatus   state;
};

struct _TfAppletClass {
	PanelAppletClass parent_class;
};

static void     tf_applet_init                 (TfApplet          *applet);
static void     tf_applet_class_init           (TfAppletClass     *klass);
static void     tf_applet_destroy              (GtkObject         *widget);

static void     tf_applet_start_stop_daemon_cb (BonoboUIComponent *uic,
						TfApplet          *applet,
						const gchar       *cname);
static void     tf_applet_about_cb             (BonoboUIComponent *uic,
						TfApplet          *applet,
						const gchar       *cname);
static void     tf_applet_change_background    (PanelApplet       *pa,
						PanelAppletBackgroundType type,
						GdkColor          *color,
						GdkPixmap         *pixmap);
static gboolean tf_applet_factory              (TfApplet          *applet,
						const gchar       *iid,
						gpointer           user_data);

static const BonoboUIVerb tf_applet_menu_verbs[] = {
	BONOBO_UI_UNSAFE_VERB ("TfAppletStartTF",
			       tf_applet_start_stop_daemon_cb),
	BONOBO_UI_UNSAFE_VERB ("TfAppletStopTF",
			       tf_applet_start_stop_daemon_cb),
	BONOBO_UI_UNSAFE_VERB ("TfAppletAbout",
			       tf_applet_about_cb),
	BONOBO_UI_VERB_END
};

G_DEFINE_TYPE (TfApplet, tf_applet, PANEL_TYPE_APPLET)

static void
tf_applet_init (TfApplet *applet)
{
	applet->image = gtk_image_new ();
	applet->tips = gtk_tooltips_new ();

	applet->state = TF_DAEMON_UNKNOWN;
	applet->source_id = 0;

	gtk_container_add (GTK_CONTAINER (applet), applet->image);
	gtk_widget_show (applet->image);
}

static void
tf_applet_class_init (TfAppletClass *klass)
{
	PanelAppletClass *applet_class = PANEL_APPLET_CLASS (klass);
	GtkObjectClass   *gtkobject_class = GTK_OBJECT_CLASS (klass);

	applet_class->change_background = tf_applet_change_background;

	gtkobject_class->destroy = tf_applet_destroy;
}

static void
tf_applet_destroy (GtkObject *widget)
{
	TfApplet *applet;

	applet = TF_APPLET (widget);

	if (applet->tips) {
		gtk_object_destroy (GTK_OBJECT (applet->tips));
		applet->tips = NULL;
	}

	GTK_OBJECT_CLASS (tf_applet_parent_class)->destroy (widget);
}

static void
tf_applet_change_background (PanelApplet *pa,
			     PanelAppletBackgroundType type,
			     GdkColor *color, GdkPixmap *pixmap)
{
	TfApplet *applet;
	/* Taken from TrashApplet */
	GtkRcStyle    *rc_style;
	GtkStyle      *style;

	applet = TF_APPLET (pa);

	/* reset style */
	gtk_widget_set_style (GTK_WIDGET (applet), NULL);
	rc_style = gtk_rc_style_new ();
	gtk_widget_modify_style (GTK_WIDGET (applet), rc_style);
	gtk_rc_style_unref (rc_style);

	switch (type) {
	case PANEL_PIXMAP_BACKGROUND:
		style = gtk_style_copy (GTK_WIDGET (applet)->style);
		if (style->bg_pixmap[GTK_STATE_NORMAL])
			g_object_unref (
				style->bg_pixmap[GTK_STATE_NORMAL]);
		style->bg_pixmap[GTK_STATE_NORMAL] = g_object_ref (
			pixmap);
		gtk_widget_set_style (GTK_WIDGET (applet), style);
		g_object_unref (style);
		break;
	case PANEL_COLOR_BACKGROUND:
		gtk_widget_modify_bg (GTK_WIDGET (applet),
				      GTK_STATE_NORMAL, color);
		break;
	case PANEL_NO_BACKGROUND:
	default:
		break;
	}
}

static void
tf_applet_menu_update_visibility (TfApplet *applet)
{
	BonoboUIComponent *popup_component;

	popup_component = panel_applet_get_popup_component (PANEL_APPLET (applet));
	
	bonobo_ui_component_set_prop (popup_component,
				      "/commands/TfAppletStartTF",
				      "hidden",
				      (applet->state == TF_DAEMON_RUNNING) ? "1" : "0",
				      NULL);
	bonobo_ui_component_set_prop (popup_component,
				      "/commands/TfAppletStopTF",
				      "hidden",
				      (applet->state == TF_DAEMON_STOPPED) ? "1" : "0",
				      NULL);
}

static void
tf_applet_start_stop_daemon_cb (BonoboUIComponent *uic,
				TfApplet          *applet,
				const gchar       *cname)
{
	switch (applet->state) {
	case TF_DAEMON_RUNNING:
		tf_daemon_stop ();
		break;
	case TF_DAEMON_STOPPED:
		tf_daemon_start ();
		break;
	default:
		break;
	}

	applet->state = tf_daemon_get_status ();
	tf_applet_menu_update_visibility (applet);
}

static void
tf_applet_about_cb (BonoboUIComponent *uic,
		    TfApplet          *applet,
		    const gchar       *cname)
{
	static const gchar *const authors[] = {
		"Carlos Garcia Campos <carlosgc@gsyc.escet.urjc.es>",
        "Edmundo Álvarez Jiménez <e.alvarezj@alumnos.urjc.es>",
		NULL
	};

	gtk_show_about_dialog (NULL,
			       "name",      "TempusFugit control applet",
			       "version",   VERSION,
			       "copyright", "\xC2\xA9 2006 Carlos Garcia Campos\n \xC2\xA9 2010 Edmundo Álvarez Jiménez",
			       "comments",  "TempusFugit daemon control utility",
			       "authors",       authors,
			       NULL);
}

static void
tf_applet_update (TfApplet *applet)
{
	static GtkIconTheme *icon_theme = NULL;
	GdkPixbuf           *pixbuf = NULL;
	guint                size = 24; /* FIXME */
	const gchar         *icon_name;
	gchar               *text_tip;
	GError              *error = NULL;

	if (!icon_theme)
		icon_theme = gtk_icon_theme_get_default ();

	/* FIXME: icon size based on panel size */
/*	size = panel_applet_get_size (PANEL_APPLET (applet));*/
	
	switch (applet->state) {
	case TF_DAEMON_UNKNOWN:
		icon_name = GTK_STOCK_STOP;
		text_tip = g_strdup ("Unknown state");
		break;
	case TF_DAEMON_RUNNING:
		icon_name = GTK_STOCK_YES;
		text_tip = g_strdup ("TempusFugit is running");
		break;
	case TF_DAEMON_STOPPED:
		icon_name = GTK_STOCK_NO;
		text_tip = g_strdup ("TempusFugit is not running");
		break;
	default:
		g_assert_not_reached ();
	}

	pixbuf = gtk_icon_theme_load_icon (icon_theme,
					   icon_name,
					   size, 0,
					   &error);

	if (error) {
		g_warning ("%s\n", error->message);
		g_error_free (error);

		return;
	}

	gtk_image_set_from_pixbuf (GTK_IMAGE (applet->image),
				   pixbuf);
	g_object_unref (pixbuf);

	gtk_widget_show (applet->image);

	gtk_tooltips_set_tip (applet->tips, GTK_WIDGET (applet), text_tip, NULL);
	g_free (text_tip);
}

static gboolean
tf_applet_run (TfApplet *applet)
{
	TfDaemonStatus state;

	state = tf_daemon_get_status ();
	if (state != applet->state) {
		applet->state = state;
		tf_applet_update (applet);
		tf_applet_menu_update_visibility (applet);
	}

	return TRUE;
}

static gboolean
tf_applet_factory (TfApplet *applet, const gchar *iid, gpointer gdata)
{
	gboolean retval = FALSE;

	if (!g_ascii_strcasecmp (iid, "OAFIID:GNOME_TfApplet")) {
		AtkObject         *atk_obj;

		panel_applet_setup_menu_from_file (PANEL_APPLET (applet),
						   DATADIR,
						   "GNOME_TfApplet.xml",
						   NULL,
						   tf_applet_menu_verbs,
						   applet);

		atk_obj = gtk_widget_get_accessible (GTK_WIDGET (applet));

		if (GTK_IS_ACCESSIBLE (atk_obj)) {
			atk_object_set_name (atk_obj, "TempusFugit control applet");
			atk_object_set_description (atk_obj, "TempusFugit daemon control utility");
		}

		applet->source_id = g_timeout_add (1000,
						   (GSourceFunc) tf_applet_run,
						   (gpointer) applet);

		gtk_widget_show (GTK_WIDGET (applet));

		retval = TRUE;
	}

	return retval;
}

PANEL_APPLET_BONOBO_FACTORY ("OAFIID:GNOME_TfApplet_Factory",
			     TF_TYPE_APPLET,
			     "tf-gnome-applet",
			     VERSION,
			     (PanelAppletFactoryCallback) tf_applet_factory,
			     NULL)

