/*
 *  Copyright (C) 2006 Carlos Garcia Campos
 *  Copyright (C) 2010 Edmundo Álvarez Jiménez
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Authors : Carlos Garcia Campos <carlosgc@gsyc.escet.urjc.es>
 *            Edmundo Álvarez Jiménez <e.alvarezj@gmail.com>
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>	
#include <stdlib.h>
#include <glib/gprintf.h>
#include <glib/gstdio.h>

#include "tf-application.h"
#include "tf-sm.h"
#include "tf-observer.h"

struct _TfApplicationPrivate {
	TfSm       *sm;
	GMainLoop  *main_loop;
	TfObserver *observer;

	gchar      *run_file;
};

#define TF_APP (tf_application_get_instance ())

#define TF_APPLICATION_GET_PRIVATE(object) \
        (G_TYPE_INSTANCE_GET_PRIVATE ((object), TF_TYPE_APPLICATION, TfApplicationPrivate))

static GObjectClass *parent_class = NULL;

/* Command line options */
static gboolean verbose_mode = FALSE;
static gboolean debug_mode = FALSE;
static gboolean version = FALSE;
static gchar *events_file = NULL;

static GOptionEntry option_entries[] = {
    { "file", 'f', 0, G_OPTION_ARG_FILENAME, &events_file,
            "Save events to FILE", "FILE" },
    { "verbose", 'v', 0, G_OPTION_ARG_NONE, &verbose_mode, "Be verbose",
            NULL },
    { "debug", 0, 0, G_OPTION_ARG_NONE, &debug_mode, "Show debug info", NULL },
    { "version", 0, 0, G_OPTION_ARG_NONE, &version,
            "Show version information and exit", NULL },
    { NULL }
};

static void tf_application_instance_init (TfApplication      *tf_app);
static void tf_application_class_init    (TfApplicationClass *klass);
static void tf_application_finalize      (GObject            *object);

GType
tf_application_get_type (void)
{
	static GType type = 0;

	if (!type) {
		static const GTypeInfo info = {
			sizeof (TfApplicationClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) tf_application_class_init,
			NULL,
			NULL,
			sizeof (TfApplication),
			0,
			(GInstanceInitFunc) tf_application_instance_init
		};

		type = g_type_register_static (G_TYPE_OBJECT, "TfApplication",
					       &info, 0);
	}

	return type;
}

static gboolean
quit_cb (gpointer user_data)
{
	tf_application_quit ();

	return FALSE;
}

static void
tf_application_signal_handler (gint signal)
{
	switch (signal) {
	case SIGTERM:
	case SIGINT:
		g_idle_add ((GSourceFunc)quit_cb, NULL);
		break;
	}
}

static void
tf_application_instance_init (TfApplication *tf_app)
{
	struct sigaction act;
	sigset_t         empty_mask;
	
	tf_app->priv = TF_APPLICATION_GET_PRIVATE (tf_app);

	tf_app->priv->sm = NULL;
	tf_app->priv->main_loop = g_main_loop_new (NULL, FALSE);
	tf_app->priv->observer = NULL;

	tf_app->priv->run_file = NULL;

	sigemptyset (&empty_mask);
	act.sa_handler = tf_application_signal_handler;
	act.sa_mask    = empty_mask;
	act.sa_flags   = 0;
	sigaction (SIGTERM,  &act, NULL);
	sigaction (SIGINT, &act, NULL);
}

static void
tf_application_class_init (TfApplicationClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	parent_class = g_type_class_peek_parent (klass);

	g_type_class_add_private (klass, sizeof (TfApplicationPrivate));

	object_class->finalize = tf_application_finalize;
}

static void
tf_application_finalize (GObject *object)
{
	TfApplication *tf_app = TF_APPLICATION (object);

	if (tf_app->priv->observer) {
		g_object_unref (tf_app->priv->observer);
		tf_app->priv->observer = NULL;
	}

	if (tf_app->priv->sm) {
		g_object_unref (tf_app->priv->sm);
		tf_app->priv->sm = NULL;
	}

	if (tf_app->priv->main_loop) {
		if (g_main_loop_is_running (tf_app->priv->main_loop)) {
			g_main_loop_quit (tf_app->priv->main_loop);
		}
		g_main_loop_unref (tf_app->priv->main_loop);
		tf_app->priv->main_loop = NULL;
	}

	if (tf_app->priv->run_file) {
		g_free (tf_app->priv->run_file);
		tf_app->priv->run_file = NULL;
	}

	if (G_OBJECT_CLASS (parent_class)->finalize)
		(* G_OBJECT_CLASS (parent_class)->finalize) (object);
}

TfApplication *
tf_application_get_instance (void)
{
	static TfApplication *tf_app = NULL;

	if (!tf_app) {
		tf_app = TF_APPLICATION (g_object_new (TF_TYPE_APPLICATION, NULL));
	}

	return tf_app;
}

void
tf_application_init (gint *argc, gchar ***argv, GError **error)
{
	TfApplication   *tf_app = TF_APP;
	GOptionContext  *context;
	static gboolean initted = FALSE;
	gchar           *command;
	
	if (initted)
		return;

	initted = TRUE;

	command = g_strdup (*argv[0]);
	
	context = g_option_context_new ("TempusFugit version " VERSION);
	g_option_context_add_main_entries (context, option_entries, NULL);
	g_option_context_parse (context, argc, argv, error);
	
	tf_app->priv->sm = tf_sm_new (command);
	tf_sm_init (tf_app->priv->sm);

	g_free (command);
	
	g_option_context_free (context);
}

/* Print debug messages if debug mode is enabled */
void
debug (const gchar *format,
    ...)
{
    va_list args;
    static gboolean debug_init = FALSE;
    static gboolean debug = TRUE;

    if (!debug)
        return;

    if (!debug_init) {
        debug = debug_mode || g_getenv ("TEMPUSFUGIT_DEBUG");

        debug_init = TRUE;

        if (!debug)
            return;
    }

	g_printerr ("DEBUG:\t");

	va_start (args, format);
	g_vfprintf (stderr, format, args);
	va_end (args);
}

void
verbose (const gchar *format,
	 ...)
{
	va_list args;
	static gboolean verbose_initted = FALSE;
	static gboolean verbose = TRUE;

	if (!verbose)
		return;
	
	if (!verbose_initted) {
		verbose = verbose_mode || g_getenv ("TEMPUSFUGIT_VERBOSE");

		verbose_initted = TRUE;

		if (!verbose)
			return;
	}

	g_printerr ("%lu:\t", (unsigned long) getpid ());

	va_start (args, format);
	g_vfprintf (stderr, format, args);
	va_end (args);
}


static gboolean
tf_application_register (TfApplication *tf_app, GError **error)
{
	FILE  *fd;
	gchar *run_dir;
	gchar *run_filename;
	gchar *run_file;

	run_dir = g_build_filename (g_get_tmp_dir (), ".detras", NULL);
	run_filename = g_strdup_printf ("tempusfugit-%s.pid", g_get_user_name ());
	run_file = g_build_filename (run_dir, run_filename, NULL);
	g_free (run_filename);

	tf_app->priv->run_file = run_file;

    debug ("run file '%s'\n", run_file);

	if (g_file_test (run_file, G_FILE_TEST_IS_REGULAR)) {
		g_set_error (error,
			     G_FILE_ERROR,
			     0,
			     "TempusFugit already running, exiting...");
		
		g_free (run_dir);

		return FALSE;
	}
	
	if (!g_file_test (run_dir, G_FILE_TEST_IS_DIR)) {
		if (g_mkdir (run_dir, 01777) == -1) {
			g_set_error (error,
				     G_FILE_ERROR,
				     1,
				     "Error while creating directory '%s'",
				     run_dir);
			
			g_free (run_dir);
			
			return FALSE;
		}

		g_chmod (run_dir, 01777);
	}

	if ((fd = g_fopen (run_file, "w")) == NULL) {
		g_set_error (error,
			     G_FILE_ERROR,
			     2,
			     "Error while creating run file '%s'",
			     run_file);
		
		g_free (run_dir);
		
		return FALSE;
	}

	g_fprintf (fd, "%u", getpid ());

	fclose (fd);
	g_free (run_dir);
	
	return TRUE;
}

static void
tf_application_unregister (TfApplication *tf_app)
{
	if (tf_app->priv->run_file &&
	    g_file_test (tf_app->priv->run_file,
			 G_FILE_TEST_IS_REGULAR)) {
		g_unlink (tf_app->priv->run_file);
	}
}

void
tf_application_run (void)
{
	TfApplication *tf_app = TF_APP;
	GError        *error = NULL;

    if (version) {
        g_printf ("TempusFugit version %s\n", VERSION);
        tf_application_quit ();
        return;
    }

    if (events_file == NULL)
        debug ("Using default events file\n");
    else
        debug ("Events file: %s\n", events_file);

	if (!tf_application_register (tf_app, &error)) {
		verbose ("%s\n", error->message);
		g_error_free (error);
		g_object_unref (tf_app);
		
		return;
	}
	
	if (g_main_loop_is_running (tf_app->priv->main_loop))
		return;

	verbose ("TempusFugit %s running\n", VERSION);
	
	/* If observer already running do nothing */
	if (!tf_app->priv->observer)
		tf_app->priv->observer = tf_observer_new (events_file);

	g_main_loop_run (tf_app->priv->main_loop);
}

void
tf_application_quit (void)
{
	static gboolean exited = FALSE;

	if (!exited) {
		verbose ("Application quitting\n");

		tf_application_unregister (TF_APP);
		
		g_object_unref (TF_APP);
		exited = TRUE;
	}
}

void
tf_application_abort (void)
{
    tf_application_unregister (TF_APP);
    g_object_unref (TF_APP);

    exit (1);
}
