/*
 *  Copyright (C) 2006 Carlos Garcia Campos
 *  Copyright (C) 2010 Edmundo Álvarez Jiménez
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Authors: Carlos Garcia Campos <carlosgc@gsyc.escet.urjc.es>
 *           Edmundo Álvarez Jiménez <e.alvarezj@gmail.com>
 */

#include <glib.h>
#include "tf-event-wc.h"

enum {
	PROP_0,
	PROP_XID,
	PROP_APP_NAME,
	PROP_WIN_TITLE
};

struct _TfEventWcPrivate {
	gint   xid;
	gchar *app_name;
	gchar *win_title;
};

#define TF_EVENT_WC_GET_PRIVATE(object) \
        (G_TYPE_INSTANCE_GET_PRIVATE ((object), TF_TYPE_EVENT_WC, TfEventWcPrivate))

static TfEventClass *parent_class = NULL;

static void tf_event_wc_init         (TfEventWc      *event);
static void tf_event_wc_class_init   (TfEventWcClass *klass);
static void tf_event_wc_finalize     (GObject        *object);

static void tf_event_wc_get_property (GObject        *object,
				      guint           prop_id,
				      GValue         *value,
				      GParamSpec     *pspec);
static void tf_event_wc_set_property (GObject        *object,
				      guint           prop_id,
				      const GValue   *value,
				      GParamSpec     *pspec);

static gboolean tf_event_wc_compare  (TfEvent        *ev1,
				      TfEvent        *ev2);

GType
tf_event_wc_get_type (void)
{
	static GType type = 0;
	
	if (!type) {
		static const GTypeInfo info = {
			sizeof (TfEventWcClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) tf_event_wc_class_init,
			NULL,
			NULL,
			sizeof (TfEventWc),
			0,
			(GInstanceInitFunc) tf_event_wc_init
		};
		
		type = g_type_register_static (TF_TYPE_EVENT, "TfEventWc",
					       &info, 0);
	}
	
	return type;
}

static void
tf_event_wc_init (TfEventWc *event)
{
	event->priv = TF_EVENT_WC_GET_PRIVATE (event);

	event->priv->xid = 0;
	event->priv->app_name = NULL;
	event->priv->win_title = NULL;
}

static void
tf_event_wc_class_init (TfEventWcClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);
	TfEventClass *event_class = TF_EVENT_CLASS (klass);

	object_class->get_property = tf_event_wc_get_property;
	object_class->set_property = tf_event_wc_set_property;

	event_class->compare = tf_event_wc_compare;

	parent_class = g_type_class_peek_parent (klass);

	g_type_class_add_private (klass, sizeof (TfEventWcPrivate));

	g_object_class_install_property (object_class,
					 PROP_XID,
					 g_param_spec_int ("xid",
							   "XID",
							   "X Window Identifier",
							   0, G_MAXINT, 0,
							   G_PARAM_READWRITE |
							   G_PARAM_CONSTRUCT_ONLY));
	g_object_class_install_property (object_class,
					 PROP_APP_NAME,
					 g_param_spec_string ("app-name",
							      "AppName",
							      "The name of the application",
							      NULL,
							      G_PARAM_READWRITE |
							      G_PARAM_CONSTRUCT_ONLY));
	g_object_class_install_property (object_class,
					 PROP_WIN_TITLE,
					 g_param_spec_string ("win-title",
							      "WinTitle",
							      "The window title",
							      NULL,
							      G_PARAM_READWRITE |
							      G_PARAM_CONSTRUCT_ONLY));

	object_class->finalize = tf_event_wc_finalize;
}

static void
tf_event_wc_finalize (GObject *object)
{

	TfEventWc *event = TF_EVENT_WC (object);

	if (event->priv->app_name) {
		g_free (event->priv->app_name);
		event->priv->app_name = NULL;
	}

	if (event->priv->win_title) {
		g_free (event->priv->win_title);
		event->priv->win_title = NULL;
	}
	
	if (G_OBJECT_CLASS (parent_class)->finalize)
		(* G_OBJECT_CLASS (parent_class)->finalize) (object);
}

static void
tf_event_wc_get_property (GObject    *object,
			  guint       prop_id,
			  GValue     *value,
			  GParamSpec *pspec)
{
	TfEventWc *event = TF_EVENT_WC (object);

	switch (prop_id) {
	case PROP_XID:
		g_value_set_int (value, event->priv->xid);
		break;
	case PROP_APP_NAME:
		g_value_set_string (value, event->priv->app_name);
		break;
	case PROP_WIN_TITLE:
		g_value_set_string (value, event->priv->win_title);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
	}
}

static void
tf_event_wc_set_property (GObject       *object,
			  guint          prop_id,
			  const GValue  *value,
			  GParamSpec    *pspec)
{
	TfEventWc *event = TF_EVENT_WC (object);

	switch (prop_id) {
	case PROP_XID:
		event->priv->xid = g_value_get_int (value);
		break;
	case PROP_APP_NAME:
        g_free(event->priv->app_name);
		event->priv->app_name = g_value_dup_string (value);
		break;
	case PROP_WIN_TITLE:
        g_free(event->priv->win_title);
		event->priv->win_title = g_value_dup_string (value);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
	}
}

TfEvent *
tf_event_wc_new (guint        xid,
		 const gchar *app_name,
		 const gchar *win_title)
{
	return TF_EVENT (g_object_new (TF_TYPE_EVENT_WC,
				       "xid", xid,
				       "app-name", app_name,
				       "win-title", win_title,
				       NULL));
}

static gboolean
tf_event_wc_compare (TfEvent *ev1, TfEvent *ev2)
{
	if (!TF_IS_EVENT_WC (ev1))
		return FALSE;

	if (!TF_IS_EVENT_WC (ev2))
		return FALSE;

	/* FIXME: ATM we only see wether xid's are the same */
	return TF_EVENT_WC (ev1)->priv->xid == TF_EVENT_WC (ev2)->priv->xid;
}
