/*
 *  Copyright (C) 2006 Carlos Garcia Campos
 *  Copyright (C) 2010 Edmundo Álvarez Jiménez
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Authors: Carlos Garcia Campos <carlosgc@gsyc.escet.urjc.es>
 *           Edmundo Álvarez Jiménez <e.alvarezj@gmail.com>
 */

#include <glib.h>
#include <X11/Xutil.h>
#include <X11/extensions/scrnsaver.h>

#include "tf-application.h"
#include "tf-idle.h"
#include "tf-scrsvr.h"
#include "tf-scrsvr-gnome.h"

struct _TfIdlePrivate {
	const Display *display;
	
	guint32        timeout;
	guint          source;

	TfIdleStatus   status;
    TfScrsvr       *scrsvr;
};

enum {
	IDLE_CHANGED,
	N_SIGNALS
};

#define INACTIVE_TIMEOUT 5 * 60 /* 5 mins */
//#define INACTIVE_TIMEOUT 10 /* For debugging purpose only */

#define TF_IDLE_GET_PRIVATE(object) \
        (G_TYPE_INSTANCE_GET_PRIVATE ((object), TF_TYPE_IDLE, TfIdlePrivate))

static GObjectClass *parent_class = NULL;
static guint         signals[N_SIGNALS];

static void     tf_idle_init       (TfIdle      *idle);
static void     tf_idle_class_init (TfIdleClass *klass);
static void     tf_idle_finalize   (GObject     *object);

static gboolean tf_idle_check      (TfIdle      *idle);

GType
tf_idle_get_type (void)
{
	static GType type = 0;

	if (!type) {
		static const GTypeInfo info = {
			sizeof (TfIdleClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) tf_idle_class_init,
			NULL,
			NULL,
			sizeof (TfIdle),
			0,
			(GInstanceInitFunc) tf_idle_init
		};

		type = g_type_register_static (G_TYPE_OBJECT, "TfIdle",
					       &info, 0);
	}

	return type;
}

static void
tf_idle_init (TfIdle *idle)
{
	idle->priv = TF_IDLE_GET_PRIVATE (idle);

	idle->priv->display = NULL;
	idle->priv->timeout = INACTIVE_TIMEOUT;
	idle->priv->status = IDLE_ACTIVE;
	idle->priv->source = g_timeout_add (1000,
					    (GSourceFunc) tf_idle_check,
					    (gpointer) idle);

    idle->priv->scrsvr = TF_SCRSVR (tf_scrsvr_gnome_new ());
}

static void
tf_idle_class_init (TfIdleClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	parent_class = g_type_class_peek_parent (klass);

	g_type_class_add_private (klass, sizeof (TfIdlePrivate));

	/* Signals */
	signals[IDLE_CHANGED] =
		g_signal_new ("changed",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (TfIdleClass, changed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE, 0);

	object_class->finalize = tf_idle_finalize;
}

static void
tf_idle_finalize (GObject *object)
{
	TfIdle *idle = TF_IDLE (object);

	if (idle->priv->timeout > 0) {
		g_source_remove (idle->priv->timeout);
		idle->priv->timeout = 0;
	}

    if (idle->priv->scrsvr) {
        g_object_unref (idle->priv->scrsvr);
        idle->priv->scrsvr = NULL;
    }

	if (G_OBJECT_CLASS (parent_class)->finalize)
		(* G_OBJECT_CLASS (parent_class)->finalize) (object);
}

TfIdle *
tf_idle_new (const Display *display)
{
	TfIdle *idle;

	g_return_val_if_fail (display != NULL, NULL);

	idle = TF_IDLE (g_object_new (TF_TYPE_IDLE, NULL));
	idle->priv->display = display;

	return idle;
}

static guint32
tf_idle_get_seconds (TfIdle *idle)
{
	static gboolean          initialized = FALSE;
	static XScreenSaverInfo *ss_info = NULL;
	gint                     event_base;
	gint                     error_base;
	gint32                   seconds = 0;

	if (!initialized) {
		if (XScreenSaverQueryExtension ((Display *)idle->priv->display,
						&event_base, &error_base)) {
			ss_info = XScreenSaverAllocInfo ();
		}

		initialized = TRUE;
	}

	if (ss_info) {
		XScreenSaverQueryInfo ((Display *)idle->priv->display,
				       DefaultRootWindow (idle->priv->display),
				       ss_info);
		seconds = ss_info->idle / 1000;
	}

	return seconds;
}


static gboolean
tf_idle_check (TfIdle *idle)
{
	guint32  seconds;
    gboolean ss_active = FALSE;
	gboolean changed = FALSE;

	seconds = tf_idle_get_seconds (idle);

	if ((seconds >= idle->priv->timeout) ||
            (tf_scrsvr_is_active (idle->priv->scrsvr))) {
		/* Inactive */
		if (idle->priv->status == IDLE_ACTIVE) {
            debug ("Inactive\n");
			changed = TRUE;
			idle->priv->status = IDLE_INACTIVE;
		}
	} else {
		/* Active */
		if (idle->priv->status == IDLE_INACTIVE) {
            debug ("Active\n");
			changed = TRUE;
			idle->priv->status = IDLE_ACTIVE;
		}
	}

	if (changed)
		g_signal_emit (G_OBJECT (idle),
			       signals[IDLE_CHANGED],
			       0);

	return TRUE;
}

void
tf_idle_set_timeout (TfIdle *idle, guint32 timeout)
{
	g_return_if_fail (TF_IS_IDLE (idle));
	g_return_if_fail (timeout > 0);

	idle->priv->timeout = timeout;
}

TfIdleStatus
tf_idle_get_status (TfIdle *idle)
{
	g_return_val_if_fail (TF_IS_IDLE (idle), IDLE_ACTIVE);

	return idle->priv->status;
}
