/*
 *  Copyright (C) 2006 Carlos Garcia Campos
 *  Copyright (C) 2010 Edmundo Álvarez Jiménez
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Authors: Carlos Garcia Campos <carlosgc@gsyc.escet.urjc.es>
 *           Edmundo Álvarez Jiménez <e.alvarezj@gmail.com>
 */

#include <glib.h>

#include "tf-application.h"
#include "tf-observer.h"
#include "tf-screen.h"
#include "tf-idle.h"
#include "tf-event.h"
#include "tf-event-wc-xml.h"
#include "tf-event-it-xml.h"
#include "tf-event-session-xml.h"
#include "tf-register.h"
#include "tf-register-xml.h"
#include "tf-xutils.h"

struct _TfObserverPrivate {
    gulong      idle_handler_id;
    gulong      wc_handler_id;
	TfScreen   *screen;
	TfRegister *regstr;
	TfIdle     *idle;
	GQueue     *event_queue;
};

#define TF_OBSERVER_GET_PRIVATE(object) \
        (G_TYPE_INSTANCE_GET_PRIVATE ((object), TF_TYPE_OBSERVER, TfObserverPrivate))

static GObjectClass *parent_class = NULL;

static void     tf_observer_init                   (TfObserver        *observer);
static void     tf_observer_register_init          (TfObserver        *observer,
                            const char        *events_file);
static void     tf_observer_class_init             (TfObserverClass   *klass);
static void     tf_observer_finalize               (GObject           *object);
static void     tf_observer_dispose                (GObject           *object);

static void     tf_observer_register_event         (TfObserver        *observer);
static gboolean tf_observer_register_event_cb      (TfObserver        *observer);
static void     tf_observer_active_window_changed  (TfObserver        *observer,
						    TfScreen          *screen);
static void     tf_observer_idle_changed           (TfObserver        *observer,
						    TfIdle            *idle);
static void     tf_observer_register_session_event (TfObserver        *observer,
						    TfEventSessionType type);

GType
tf_observer_get_type (void)
{
	static GType type = 0;

	if (!type) {
		static const GTypeInfo info = {
			sizeof (TfObserverClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) tf_observer_class_init,
			NULL,
			NULL,
			sizeof (TfObserver),
			0,
			(GInstanceInitFunc) tf_observer_init
		};

		type = g_type_register_static (G_TYPE_OBJECT, "TfObserver",
					       &info, 0);
	}

	return type;
}

/*
 * Instance initializer. Make sure you are calling this method through
 * tf_observer_new(), which is responsible of initialize register.
 */
static void
tf_observer_init (TfObserver *observer)
{
	observer->priv = TF_OBSERVER_GET_PRIVATE (observer);

	observer->priv->screen = tf_screen_new ();
	observer->priv->idle = tf_idle_new (tf_screen_get_display (observer->priv->screen));

	observer->priv->wc_handler_id = g_signal_connect_swapped(
            G_OBJECT(observer->priv->screen), "active-window-changed",
            G_CALLBACK(tf_observer_active_window_changed),
            (gpointer) observer);
    
	observer->priv->idle_handler_id = g_signal_connect_swapped(
            G_OBJECT(observer->priv->idle), "changed",
            G_CALLBACK(tf_observer_idle_changed),
            (gpointer) observer);
	
	observer->priv->event_queue = g_queue_new ();

	tf_observer_register_session_event (observer, TF_EVENT_SESSION_INIT);
    tf_observer_active_window_changed (observer, observer->priv->screen);
}

static void
tf_observer_class_init (TfObserverClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	parent_class = g_type_class_peek_parent (klass);

	g_type_class_add_private (klass, sizeof (TfObserverPrivate));

	object_class->finalize = tf_observer_finalize;
	object_class->dispose = tf_observer_dispose;
}

/* Initializes register with a given events file*/
static void
tf_observer_register_init (TfObserver *observer, const char *events_file)
{
	observer->priv->regstr = TF_REGISTER (tf_register_xml_new (events_file));
}

static void
tf_observer_dispose (GObject *object)
{
	TfObserver *observer = TF_OBSERVER (object);

    /* Let's disconnect signal handlers */
    g_signal_handler_disconnect(G_OBJECT(observer->priv->screen),
            observer->priv->wc_handler_id);
    g_signal_handler_disconnect(G_OBJECT(observer->priv->idle),
            observer->priv->idle_handler_id);

	if (observer->priv->event_queue) {
		tf_observer_register_session_event (observer, TF_EVENT_SESSION_END);

		/* Flushing loop */
		while (!g_queue_is_empty (observer->priv->event_queue))
			g_main_context_iteration (NULL, FALSE);

		g_queue_free (observer->priv->event_queue);
		observer->priv->event_queue = NULL;
	}

	/* Release any possible lock */
	if (observer->priv->regstr)
		tf_register_unlock (observer->priv->regstr);

	if (G_OBJECT_CLASS (parent_class)->dispose)
		(* G_OBJECT_CLASS (parent_class)->dispose) (object);
}

static void
tf_observer_finalize (GObject *object)
{
	TfObserver *observer = TF_OBSERVER (object);

	if (observer->priv->screen) {
		g_object_unref (observer->priv->screen);
		observer->priv->screen = NULL;
	}

	if (observer->priv->regstr) {
		g_object_unref (observer->priv->regstr);
		observer->priv->regstr = NULL;
	}

	if (G_OBJECT_CLASS (parent_class)->finalize)
		(* G_OBJECT_CLASS (parent_class)->finalize) (object);
}

TfObserver *
tf_observer_new (const gchar *events_file)
{
	TfObserver *observer;

    observer = TF_OBSERVER (g_object_new (TF_TYPE_OBSERVER, NULL));
    tf_observer_register_init (observer, events_file);

	return observer;
}

static gboolean
tf_observer_register_event_cb (TfObserver *observer)
{
	TfEvent        *event;
	static TfEvent *last_event = NULL;

	if (g_queue_is_empty (observer->priv->event_queue))
		return FALSE;

	if (!tf_register_lock (observer->priv->regstr))
		return TRUE; /* Try later */

	event = g_queue_pop_head (observer->priv->event_queue);

	if (last_event) {
		if (tf_event_compare (event, last_event)) {
			g_object_unref (last_event);
			last_event = g_object_ref (event);
			g_object_unref (event);

			tf_register_unlock (observer->priv->regstr);

			return FALSE;
		}
	}

	tf_register_save_event (observer->priv->regstr, event);
	verbose ("Event registered\n");

	if (last_event)
		g_object_unref (last_event);
	last_event = g_object_ref (event);

	g_object_unref (event);

	tf_register_unlock (observer->priv->regstr);

	return FALSE;
}

static void
tf_observer_register_event (TfObserver *observer)
{
	/* Use a timeout instead of idle because it's possible
	 * that we can't register the event (resource locked).
	 * We want to try it later but after a second or so.
	 * In this way we don't saturate the cpu
	 */
	g_timeout_add_full (G_PRIORITY_LOW,
			    1000,
			    (GSourceFunc) tf_observer_register_event_cb,
			    (gpointer) observer, NULL);
}

static void
tf_observer_active_window_changed (TfObserver *observer, TfScreen *screen)
{
	Display *display;
	Window   win;
	TfEvent *event;
	gchar   *app_name;
	gchar   *win_title;

	win = tf_screen_get_active_window (observer->priv->screen);
	if (!win)
		return;
	
	display = tf_screen_get_display (observer->priv->screen);
	
	app_name = tf_get_application_name (display, win);
	win_title = tf_get_window_title (display, win);

	event = TF_EVENT (tf_event_wc_xml_new (win, app_name, win_title));
	g_queue_push_tail (observer->priv->event_queue, event);
	verbose ("Queued window_changed event (%s - %s)\n", app_name, win_title);

	g_free (app_name);
	g_free (win_title);

	tf_observer_register_event (observer);
}

static void
tf_observer_idle_changed (TfObserver *observer, TfIdle *idle)
{
	TfEvent *event;

	switch (tf_idle_get_status (idle)) {
	case IDLE_ACTIVE:
		tf_observer_active_window_changed (observer, observer->priv->screen);
		
		break;
	case IDLE_INACTIVE:
		event = TF_EVENT (tf_event_it_xml_new ());
		g_queue_push_tail (observer->priv->event_queue, event);
		verbose ("Queued idle event\n");
		
		tf_observer_register_event (observer);
		
		break;
	}
}

static void
tf_observer_register_session_event (TfObserver *observer, TfEventSessionType type)
{
	TfEvent *event;

	event = TF_EVENT (tf_event_session_xml_new (type));
	
   	g_queue_push_tail (observer->priv->event_queue, event);
	verbose ("Queued session event (%s)\n",
		 type == TF_EVENT_SESSION_INIT ? "Init" : "End");
	
	tf_observer_register_event (observer);
}
