/*
 *  Copyright (C) 2006 Carlos Garcia Campos
 *  Copyright (C) 2010 Edmundo Álvarez Jiménez
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Authors: Carlos Garcia Campos <carlosgc@gsyc.escet.urjc.es>
 *           Edmundo Álvarez Jiménez <e.alvarezj@gmail.com>
 */

#include <glib.h>
#include <glib/gstdio.h>
#include <unistd.h>

#include "tf-event-xml.h"
#include "tf-register.h"
#include "tf-register-xml.h"
#include "tf-application.h"
#include "tf-xml-validator.h"

enum {
    PROP_0,
    PROP_FILE
};

struct _TfRegisterXmlPrivate {
	gchar           *xml_file;
	GIOChannel      *xml;
    TfXmlValidator  *val;

	guint           timeout_lock_id;
};

#define TF_REGISTER_XML_GET_PRIVATE(object) \
        (G_TYPE_INSTANCE_GET_PRIVATE ((object), TF_TYPE_REGISTER_XML, TfRegisterXmlPrivate))

#define LOCK_TIMEOUT 5*60*1000 /* 5 min. */

//Default filename where events are going to be saved
#define DEFAULT_EVENTS_FILE "tempusfugit.xml"

static GObjectClass *parent_class = NULL;

static void     tf_register_xml_init          (TfRegisterXml      *reg);
static void     tf_register_xml_class_init    (TfRegisterXmlClass *klass);
static void     tf_register_xml_register_init (TfRegisterIface    *iface);
static void     tf_register_xml_finalize      (GObject            *object);

static void     tf_register_xml_init_file    (GIOChannel         *file);
static void     tf_register_xml_save_event    (TfRegister         *reg,
					       const TfEvent      *event);
static gboolean tf_register_xml_lock          (TfRegister         *reg);
static void     tf_register_xml_unlock        (TfRegister         *reg);

static void     tf_register_xml_timeout_start (TfRegisterXml      *reg);
static void     tf_register_xml_timeout_stop  (TfRegisterXml      *reg);

static void     tf_register_xml_set_property  (GObject            *object,
                            guint           prop_id,
                            const GValue    *value,
                            GParamSpec      *pspec);
					       
GType
tf_register_xml_get_type (void)
{
	static GType type = 0;

	if (!type) {
		static const GTypeInfo info = {
			sizeof (TfRegisterXmlClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) tf_register_xml_class_init,
			NULL,
			NULL,
			sizeof (TfRegisterXml),
			0,
			(GInstanceInitFunc) tf_register_xml_init
		};

		static const GInterfaceInfo register_info = {
			(GInterfaceInitFunc) tf_register_xml_register_init,
			NULL,
			NULL
		};

		type = g_type_register_static (G_TYPE_OBJECT, "TfRegisterXml",
					       &info, 0);

		g_type_add_interface_static (type, TF_TYPE_REGISTER,
					     &register_info);
	}

	return type;
}


/* Initializes a new events' file */
static void
tf_register_xml_init_file (GIOChannel *file)
{
    GError *error = NULL;

    g_io_channel_write_chars (file, "<?xml version=\"1.0\"?>\n"
            "<events>\n</events>\n", -1, NULL, &error);

    if (error) {
        g_critical ("%s", error->message);
        g_error_free (error);
        g_io_channel_unref (file);
        tf_application_abort ();
    }
}


static void
tf_register_xml_constructed (GObject *object)
{
    TfRegisterXml   *reg = TF_REGISTER_XML (object);
	gchar           *filename;
	gchar           *dirname;
    struct stat     filestat;
	GError          *error = NULL;
    GIOChannel      *new_file;
	
    debug ("Executing constructed method\n");

    if (g_strcmp0(reg->priv->xml_file, DEFAULT_EVENTS_FILE) == 0) {
        debug ("Default filename\n");
    	dirname = g_build_filename (g_get_home_dir (), ".detras", NULL);
        filename = reg->priv->xml_file;
    } else {
        debug ("Custom filename\n");
        dirname = g_path_get_dirname (reg->priv->xml_file);
        filename = g_path_get_basename (reg->priv->xml_file);

        g_free (reg->priv->xml_file);
    }

    if (!g_file_test (dirname, G_FILE_TEST_IS_DIR)) {
        gint res;

        if (g_mkdir (dirname, 0755) < 0) {
            g_critical ("Cannot create '%s' directory\n", dirname);
            tf_application_abort ();
        }
   	}

    debug ("dirname: %s\n", dirname);
    debug ("filename: %s\n", filename);

    reg->priv->xml_file = g_build_filename (dirname, filename, NULL);
    debug ("file: %s\n", reg->priv->xml_file);

    g_free (dirname);
    g_free (filename);

    /* Checks if the file is right and initializes it if it is needed */
    if (g_file_test (reg->priv->xml_file, G_FILE_TEST_EXISTS)) {
        debug ("file exists\n");

        if (g_file_test (reg->priv->xml_file, G_FILE_TEST_IS_REGULAR)) {
            debug ("regular file\n");
            new_file = g_io_channel_new_file (reg->priv->xml_file,
                    "r+", &error);
            if (error) {
                g_critical ("%s", error->message);
                g_error_free (error);
                tf_application_abort ();
            }

            if (g_stat (reg->priv->xml_file, &filestat) == -1) {
                g_critical ("cannot stat '%s'\n", reg->priv->xml_file);
                g_io_channel_unref (new_file);
                tf_application_abort ();
            }

            if (filestat.st_size == 0) {
                debug ("empty file\n");
                tf_register_xml_init_file (new_file);
            } else {
                debug ("non-empty file\n");

                reg->priv->val = tf_xml_validator_new (reg->priv->xml_file);

                if ((!reg->priv->val) || (!validate (reg->priv->val))) {
                    g_critical ("'%s' is a malformed XML file\n",
                            reg->priv->xml_file);
                    tf_application_abort ();
                }
            }
        } else {
            g_critical ("'%s' is not a regular file\n", reg->priv->xml_file);
            tf_application_abort ();
        }
    } else {
        debug ("file does not exists\n");
		
        new_file = g_io_channel_new_file (reg->priv->xml_file, "w", &error);
        if (error) {
            g_critical ("%s", error->message);
            g_error_free (error);
            tf_application_abort ();
        }

        tf_register_xml_init_file (new_file);

        g_io_channel_unref (new_file);
    }

    reg->priv->xml = g_io_channel_new_file (reg->priv->xml_file, "r+",
            &error);
    if (error) {
        g_critical ("%s", error->message);
        g_error_free (error);
        tf_application_abort ();
    }
}


static void
tf_register_xml_init (TfRegisterXml *reg)
{
	reg->priv = TF_REGISTER_XML_GET_PRIVATE (reg);
	reg->priv->xml = NULL;
	reg->priv->timeout_lock_id = 0;
}

static void
tf_register_xml_class_init (TfRegisterXmlClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->constructed = tf_register_xml_constructed;
    object_class->set_property = tf_register_xml_set_property;

	parent_class = g_type_class_peek_parent (klass);

	g_type_class_add_private (klass, sizeof (TfRegisterXmlPrivate));

    debug ("Installing property\n");
    g_object_class_install_property (object_class,
            PROP_FILE,
            g_param_spec_string ("xml_file",
                    "XML File",
                    "Name of events file",
                    DEFAULT_EVENTS_FILE,
                    G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY));
    debug ("Done!\n");

	object_class->finalize = tf_register_xml_finalize;
}


static void
tf_register_xml_set_property (GObject *object,
        guint           prop_id,
        const GValue    *value,
        GParamSpec      *pspec)
{
    debug ("Setting property\n");
    TfRegisterXml *reg = TF_REGISTER_XML (object);

    debug ("prop_id = %lu\n", prop_id);
    debug ("value = %s\n", g_value_get_string(value));

    switch (prop_id) {
    case PROP_FILE:
        g_free (reg->priv->xml_file);
        reg->priv->xml_file = g_value_dup_string (value);
        g_assert (reg->priv->xml_file != NULL);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}


static void
tf_register_xml_register_init (TfRegisterIface *iface)
{
	iface->save_event = tf_register_xml_save_event;
	iface->lock = tf_register_xml_lock;
	iface->unlock = tf_register_xml_unlock;
}

static void
tf_register_xml_finalize (GObject *object)
{
	TfRegisterXml *reg = TF_REGISTER_XML (object);

	if (reg->priv->xml) {
		g_io_channel_unref (reg->priv->xml);
		reg->priv->xml = NULL;
	}

	if (reg->priv->xml_file) {
		g_free (reg->priv->xml_file);
		reg->priv->xml_file = NULL;
	}

    if (reg->priv->val) {
        g_object_unref (reg->priv->val);
        reg->priv->val = NULL;
    }

	tf_register_xml_timeout_stop (reg);
	
	if (G_OBJECT_CLASS (parent_class)->finalize)
		(* G_OBJECT_CLASS (parent_class)->finalize) (object);
}

TfRegisterXml *
tf_register_xml_new (const gchar *events_file)
{
	TfRegisterXml *reg;

    debug ("New TfRegisterXml\n");
    if (events_file == NULL)
        reg = TF_REGISTER_XML (g_object_new (TF_TYPE_REGISTER_XML, NULL));
    else
    	reg = TF_REGISTER_XML (g_object_new (TF_TYPE_REGISTER_XML, 
                "xml_file", events_file, NULL));
    debug ("Created!\n");

    if (reg == NULL)
        debug ("NULL!\n");
    else
        debug ("NOT NULL\n");

	return reg;
}

static void
tf_register_xml_save_event (TfRegister *reg, const TfEvent *event)
{
	TfRegisterXml *regstr = TF_REGISTER_XML (reg);
	TfEventXml    *evxml = TF_EVENT_XML (event);
	gchar         *event_xml, *xml;
	GError        *error = NULL;

	event_xml = tf_event_xml_get_xml (evxml);
	xml = g_strdup_printf ("%s</events>\n", event_xml);
	g_free (event_xml);

	g_io_channel_seek_position (regstr->priv->xml,
				    g_utf8_strlen ("</events>\n", -1) * -1,
				    G_SEEK_END, &error);
	if (error) {
		g_warning ("%s", error->message);
		g_error_free (error);
	    error = NULL;
	}

	g_io_channel_write_chars (regstr->priv->xml, xml, 
				  -1, NULL, &error);
    if (error) {
        g_warning ("%s", error->message);
        g_error_free (error);
	    error = NULL;
    }

	g_io_channel_flush (regstr->priv->xml, &error);
    if (error) {
        g_warning ("%s", error->message);
        g_error_free (error);
    }
	
	g_free (xml);
}

static gboolean
tf_register_xml_timeout (TfRegisterXml *reg)
{
	verbose ("Lock timeout reached, releasing lock\n");
	tf_register_xml_unlock (TF_REGISTER (reg));

	return FALSE;
}

static void
tf_register_xml_timeout_start (TfRegisterXml *reg)
{
	if (reg->priv->timeout_lock_id == 0) {
		verbose ("Lock timeout started\n");
		reg->priv->timeout_lock_id =
			g_timeout_add (LOCK_TIMEOUT,
				       (GSourceFunc) tf_register_xml_timeout,
				       reg);
	}
}

static void
tf_register_xml_timeout_stop (TfRegisterXml *reg)
{
	if (reg->priv->timeout_lock_id > 0) {
		verbose ("Lock timeout stopped\n");
		g_source_remove (reg->priv->timeout_lock_id);
		reg->priv->timeout_lock_id = 0;
	}
}

static gboolean
tf_register_xml_lock (TfRegister *reg)
{
	TfRegisterXml *regstr = TF_REGISTER_XML (reg);
	gchar         *lock_file;

	lock_file = g_strdup_printf ("%s.lock",
				     regstr->priv->xml_file);

	verbose ("Getting lock %s\n", lock_file);
	if (link (regstr->priv->xml_file, lock_file) == -1) {
		verbose ("File %s is locked\n", lock_file);
		g_free (lock_file);

		tf_register_xml_timeout_start (regstr);

		return FALSE;
	}

	tf_register_xml_timeout_stop (regstr);

	verbose ("Locked %s\n", lock_file);
	g_free (lock_file);
	
	return TRUE;	
}

static void
tf_register_xml_unlock (TfRegister *reg)
{
	TfRegisterXml *regstr = TF_REGISTER_XML (reg);
	gchar         *lock_file;
	
	lock_file = g_strdup_printf ("%s.lock",
				     regstr->priv->xml_file);

	verbose ("Unlock %s\n", lock_file);
	unlink (lock_file);
	
	g_free (lock_file);
}
