/*
 *  Copyright (C) 2006 Carlos Garcia Campos
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Authors : Carlos Garcia Campos <carlosgc@gsyc.escet.urjc.es>
 */

#include <glib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <stdlib.h>

#include "tf-application.h"
#include "tf-screen.h"
#include "tf-xutils.h"

typedef struct _TfScreenSource TfScreenSource;

struct _TfScreenPrivate {
	Display *display;

	Window   active_window;
	gchar   *active_window_title;
};

struct _TfScreenSource {
	GSource   source;

	TfScreen *screen;
	GPollFD   event_poll_fd;
};

enum {
	ACTIVE_WINDOW_CHANGED,
	N_SIGNALS
};

#define TF_SCREEN_GET_PRIVATE(object) \
        (G_TYPE_INSTANCE_GET_PRIVATE ((object), TF_TYPE_SCREEN, TfScreenPrivate))

static GObjectClass *parent_class = NULL;
static guint         signals[N_SIGNALS];

static void     tf_screen_init           (TfScreen      *screen);
static void     tf_screen_class_init     (TfScreenClass *klass);
static void     tf_screen_finalize       (GObject       *object);

/* Screen Source */
static GSource *tf_screen_source_new     (TfScreen      *screen);
static gboolean tf_screen_event_prepare  (GSource       *source,
					  gint          *timeout);
static gboolean tf_screen_event_check    (GSource       *source);
static gboolean tf_screen_event_dispatch (GSource       *source,
					  GSourceFunc    callback,
					  gpointer       user_data);

GType
tf_screen_get_type (void)
{
	static GType type = 0;

	if (!type) {
		static const GTypeInfo info = {
			sizeof (TfScreenClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) tf_screen_class_init,
			NULL,
			NULL,
			sizeof (TfScreen),
			0,
			(GInstanceInitFunc) tf_screen_init
		};

		type = g_type_register_static (G_TYPE_OBJECT, "TfScreen",
					       &info, 0);
	}

	return type;
}

static void
tf_screen_init (TfScreen *screen)
{
	GSource          *source;
	TfScreenSource   *screen_source;
	Window            root_window;
	XWindowAttributes attr;

	screen->priv = TF_SCREEN_GET_PRIVATE (screen);

	screen->priv->display = XOpenDisplay (NULL);
	g_assert (screen->priv->display != NULL);

	tf_x_error_init ();

	screen->priv->active_window = None;
	screen->priv->active_window_title = NULL;
	
	root_window = DefaultRootWindow (screen->priv->display);

	XGetWindowAttributes (screen->priv->display,
			      root_window, &attr);
	XSelectInput (screen->priv->display,
		      root_window,
		      attr.your_event_mask | PropertyChangeMask);
	
	source = tf_screen_source_new (screen);
	screen_source = (TfScreenSource *) source;
	
	g_source_set_priority (source, G_PRIORITY_DEFAULT);
	g_source_add_poll (source, &screen_source->event_poll_fd);
	g_source_set_can_recurse (source, TRUE);
	g_source_attach (source, NULL);
}

static void
tf_screen_class_init (TfScreenClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	parent_class = g_type_class_peek_parent (klass);

	g_type_class_add_private (klass, sizeof (TfScreenPrivate));

	/* Signals */
	signals[ACTIVE_WINDOW_CHANGED] =
		g_signal_new ("active_window_changed",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (TfScreenClass, active_window_changed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE, 0);

	object_class->finalize = tf_screen_finalize;
}

static void
tf_screen_finalize (GObject *object)
{
	TfScreen *screen = TF_SCREEN (object);

	if (screen->priv->display) {
		XCloseDisplay (screen->priv->display);
		screen->priv->display = NULL;
	}

	if (screen->priv->active_window_title) {
		g_free (screen->priv->active_window_title);
		screen->priv->active_window_title = NULL;
	}

	if (G_OBJECT_CLASS (parent_class)->finalize)
		(* G_OBJECT_CLASS (parent_class)->finalize) (object);
}

TfScreen *
tf_screen_new (void)
{
	TfScreen *screen;

	screen = TF_SCREEN (g_object_new (TF_TYPE_SCREEN, NULL));

	return screen;
}

Window
tf_screen_get_active_window (TfScreen *screen)
{
	g_return_val_if_fail (TF_IS_SCREEN (screen), None);
	
	return tf_get_active_window (screen->priv->display);
}

Display *
tf_screen_get_display (TfScreen *screen)
{
	g_return_val_if_fail (TF_IS_SCREEN (screen), NULL);
	
	return screen->priv->display;
}

static GSourceFuncs event_funcs = {
	tf_screen_event_prepare,
	tf_screen_event_check,
	tf_screen_event_dispatch,
	NULL
};

static GSource *
tf_screen_source_new (TfScreen *screen)
{
	GSource        *source;
	TfScreenSource *screen_source;

	source = g_source_new (&event_funcs, sizeof (TfScreenSource));
	screen_source = (TfScreenSource *)source;
	
	screen_source->screen = screen;
	screen_source->event_poll_fd.fd = ConnectionNumber (screen->priv->display);
	screen_source->event_poll_fd.events = G_IO_IN;

	return source;
}

static gboolean
tf_screen_event_prepare (GSource  *source,
			 gint     *timeout)
{
	TfScreen *screen = ((TfScreenSource *)source)->screen;
	gboolean  retval;

	*timeout = -1;
	retval = XPending (screen->priv->display);

	return retval;
}

static gboolean
tf_screen_event_check (GSource *source)
{
	TfScreenSource *screen_source = (TfScreenSource *)source;
	TfScreen       *screen = ((TfScreenSource *)source)->screen;
	gboolean        retval;

	if (screen_source->event_poll_fd.revents & G_IO_IN)
		retval = XPending (screen->priv->display);
	else
		retval = FALSE;

	return retval;
}

static gboolean
tf_screen_active_window_changed (TfScreen *screen)
{
	Window window = None;

	window = tf_get_active_window (screen->priv->display);
	if (window == None)
		return FALSE;

	if (screen->priv->active_window != window) {
		screen->priv->active_window = window;
		
		if (screen->priv->active_window_title)
			g_free (screen->priv->active_window_title);
		
		screen->priv->active_window_title =
				tf_get_window_title (screen->priv->display,
						     window);
		return TRUE;
	} else {
		gchar *title;

		title = tf_get_window_title (screen->priv->display, window);
		
		if (screen->priv->active_window_title && title &&
		    g_ascii_strcasecmp (title, screen->priv->active_window_title) != 0) {
			g_free (screen->priv->active_window_title);
			screen->priv->active_window_title = title;

			return TRUE;
		}

		g_free (title);
	}
	
	return FALSE;
}

static gboolean
tf_screen_event_dispatch (GSource    *source,
			  GSourceFunc callback,
			  gpointer    user_data)
{
	TfScreen *screen = ((TfScreenSource *)source)->screen;
	XEvent     event;
	
	while (XPending (screen->priv->display)) {
		XNextEvent (screen->priv->display, &event);

		switch (event.type) {
		case PropertyNotify:
			if (event.xproperty.atom == _active_window_atom (screen->priv->display)) {
				if (tf_screen_active_window_changed (screen))
					g_signal_emit (G_OBJECT (screen),
						       signals[ACTIVE_WINDOW_CHANGED],
						       0);
			}
			break;
		default:
			continue;
		}
	}

	return TRUE;
}
