/*
 *  Copyright (C) 2010 Edmundo Álvarez Jiménez
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Authors: Edmundo Álvarez Jiménez <e.alvarezj@gmail.com>
 */

#include <glib.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-bindings.h>

#include "tf-scrsvr.h"
#include "tf-scrsvr-gnome.h"

struct _TfScrsvrGnomePrivate {
    DBusGConnection *connection;
    DBusGProxy      *proxy;
};

#define GNOME_SCRSVR_SERVICE "org.gnome.ScreenSaver"
#define GNOME_SCRSVR_PATH "/org/gnome/ScreenSaver"
#define GNOME_SCRSVR_IFACE "org.gnome.ScreenSaver"

#define TF_SCRSVR_GNOME_GET_PRIVATE(object) \
        (G_TYPE_INSTANCE_GET_PRIVATE ((object), TF_TYPE_SCRSVR_GNOME, TfScrsvrGnomePrivate))

static GObjectClass *parent_class = NULL;

static void     tf_scrsvr_gnome_init        (TfScrsvrGnome      *ssg);
static void     tf_scrsvr_gnome_class_init  (TfScrsvrGnomeClass *klass);
static void     tf_scrsvr_gnome_scrsvr_init (TfScrsvrIface      *iface);
static void     tf_scrsvr_gnome_finalize    (GObject            *object);

static gboolean tf_scrsvr_gnome_is_active   (TfScrsvr           *ss);


GType
tf_scrsvr_gnome_get_type (void)
{
	static GType type = 0;
	
	if (!type) {
		static const GTypeInfo info = {
			sizeof (TfScrsvrGnomeClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) tf_scrsvr_gnome_class_init,
			NULL,
			NULL,
			sizeof (TfScrsvrGnome),
			0,
			(GInstanceInitFunc) tf_scrsvr_gnome_init
		};

		static const GInterfaceInfo register_info = {
			(GInterfaceInitFunc) tf_scrsvr_gnome_scrsvr_init,
			NULL,
			NULL
		};
		
		type = g_type_register_static (G_TYPE_OBJECT, "TfScrsvrGnome",
					       &info, 0);
		
		g_type_add_interface_static (type, TF_TYPE_SCRSVR,
					     &register_info);
	}
	
	return type;
}


static void
tf_scrsvr_gnome_init (TfScrsvrGnome *ssg)
{
    GError *error = NULL;

    ssg->priv = TF_SCRSVR_GNOME_GET_PRIVATE (ssg);

    ssg->priv->connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
    if (error) {
        g_warning ("Failed to connect to bus: %s", error->message);
        g_error_free (error);
        error = NULL;
    }

    ssg->priv->proxy = dbus_g_proxy_new_for_name (ssg->priv->connection,
            GNOME_SCRSVR_SERVICE, GNOME_SCRSVR_PATH, GNOME_SCRSVR_IFACE);
}


static void
tf_scrsvr_gnome_class_init (TfScrsvrGnomeClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    parent_class = g_type_class_peek_parent (klass);

    g_type_class_add_private (klass, sizeof (TfScrsvrGnomePrivate));

    object_class->finalize = tf_scrsvr_gnome_finalize;
}


static void
tf_scrsvr_gnome_finalize (GObject *object)
{
    TfScrsvrGnome *ssg = TF_SCRSVR_GNOME (object);

    if (ssg->priv->proxy) {
        g_object_unref (ssg->priv->proxy);
        ssg->priv->proxy = NULL;
    }

    if (ssg->priv->connection) {
        dbus_g_connection_unref (ssg->priv->connection);
        ssg->priv->connection = NULL;
    }

    if (G_OBJECT_CLASS (parent_class)->finalize)
        (* G_OBJECT_CLASS (parent_class)->finalize) (object);
}


static void
tf_scrsvr_gnome_scrsvr_init (TfScrsvrIface *iface)
{
	iface->is_active = tf_scrsvr_gnome_is_active;
}


TfScrsvrGnome *
tf_scrsvr_gnome_new (void)
{
    TfScrsvrGnome *ssg;

    ssg = TF_SCRSVR_GNOME (g_object_new (TF_TYPE_SCRSVR_GNOME, NULL));

    return ssg;
}


static gboolean
tf_scrsvr_gnome_is_active (TfScrsvr *ss)
{
    TfScrsvrGnome *ssg = TF_SCRSVR_GNOME (ss);
    GError        *error = NULL;
    gboolean      response = FALSE;

    if (!ssg->priv->proxy)
        return FALSE;

    if (!dbus_g_proxy_call (ssg->priv->proxy, "GetActive", &error,
            G_TYPE_INVALID, G_TYPE_BOOLEAN, &response, G_TYPE_INVALID)) {
        g_error_free (error);
        return FALSE;
    }
    
    return response;
}

