/*
 *  Copyright (C) 2006 Carlos Garcia Campos
 *  Copyright (C) 2010 Edmundo Álvarez Jiménez
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Authors: Carlos Garcia Campos <carlosgc@gsyc.escet.urjc.es>
 *           Edmundo Álvarez Jiménez <e.alvarezj@gmail.com>
 */

#include <glib.h>
#include <glib/gstdio.h>
#include <X11/ICE/ICElib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>

#include "tf-application.h"
#include "tf-xutils.h"
#include "tf-sm.h"

struct _TfSmPrivate {
	SmcConn  conn;
	gchar   *command;
};

enum {
	PROP0,
	PROP_COMMAND
};

#define TF_SM_GET_PRIVATE(object) \
        (G_TYPE_INSTANCE_GET_PRIVATE ((object), TF_TYPE_SM, TfSmPrivate))

#define TF_APP_NAME "TempusFugit"

static GObjectClass *parent_class = NULL;

static void tf_sm_instance_init (TfSm         *sm);
static void tf_sm_class_init    (TfSmClass    *klass);
static void tf_sm_finalize      (GObject      *object);

static void tf_sm_set_property  (GObject      *object,
				 guint         prop_id,
				 const GValue *value,
				 GParamSpec   *pspec);

GType
tf_sm_get_type (void)
{
	static GType type = 0;

	if (!type) {
		static const GTypeInfo info = {
			sizeof (TfSmClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) tf_sm_class_init,
			NULL,
			NULL,
			sizeof (TfSm),
			0,
			(GInstanceInitFunc) tf_sm_instance_init
		};

		type = g_type_register_static (G_TYPE_OBJECT, "TfSm",
					       &info, 0);
	}

	return type;
}

static void
tf_sm_instance_init (TfSm *sm)
{
	sm->priv = TF_SM_GET_PRIVATE (sm);

	sm->priv->conn = NULL;
	sm->priv->command = NULL;
}

static void
tf_sm_class_init (TfSmClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	parent_class = g_type_class_peek_parent (klass);

	object_class->set_property = tf_sm_set_property;

	g_type_class_add_private (klass, sizeof (TfSmPrivate));

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_COMMAND,
					 g_param_spec_string ("command",
							      "Command",
							      "The name of the command executed",
							      NULL,
							      G_PARAM_WRITABLE |
							      G_PARAM_CONSTRUCT_ONLY));
	
	object_class->finalize = tf_sm_finalize;
}

static void
tf_sm_finalize (GObject *object)
{
	TfSm *sm = TF_SM (object);

	if (sm->priv->command) {
		g_free (sm->priv->command);
		sm->priv->command = NULL;
	}
	
	if (sm->priv->conn) {
		SmcCloseConnection (sm->priv->conn, 0, NULL);
		sm->priv->conn = NULL;
	}

	if (G_OBJECT_CLASS (parent_class)->finalize)
		(* G_OBJECT_CLASS (parent_class)->finalize) (object);
}

static void
tf_sm_set_property (GObject      *object,
		    guint         prop_id,
		    const GValue *value,
		    GParamSpec   *pspec)
{
	TfSm *sm = TF_SM (object);

	switch (prop_id) {
	case PROP_COMMAND:
		sm->priv->command = g_value_dup_string (value);
		g_assert (sm->priv->command != NULL);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
	}
}

TfSm *
tf_sm_new (const gchar *command)
{
	TfSm *sm;

	g_return_val_if_fail (command != NULL, NULL);
	
	sm = TF_SM (g_object_new (TF_TYPE_SM,
				  "command", command,
				  NULL));

	return sm;
}

/* Session callbacks */
static void
save_yourself_callback (SmcConn   smc_conn,
			SmPointer client_data,
			int       save_style,
			gboolean  shutdown,
			int       interact_style,
			gboolean  fast)
{
	SmcSaveYourselfDone (smc_conn, True);
}

static void
save_complete_callback (SmcConn   smc_conn,
			SmPointer client_data)
{
	/* Nothing to do */
}

static void
shutdown_cancelled_callback (SmcConn   smc_conn,
			     SmPointer client_data)
{
	/* Nothing to do */
}

static void
die_callback (SmcConn   smc_conn,
	      SmPointer user_data)
{
	tf_application_quit ();
}

/* ICE messages processing (stolen from gnome-client) */
static IceIOErrorHandler ice_installed_handler;

static gboolean
process_ice_messages (GIOChannel   *source,
		      GIOCondition  condition,
		      gpointer      data)
{
	IceConn conn = (IceConn) data;
	IceProcessMessagesStatus status;

	status = IceProcessMessages (conn, NULL, NULL);

	if (status == IceProcessMessagesIOError) {
		IceSetShutdownNegotiation (conn, False);
		IceCloseConnection (conn);
	}
	
	return TRUE;
}

static void
new_ice_connection (IceConn     conn,
		    IcePointer  client_data,
		    Bool        opening,
		    IcePointer *watch_data)
{
	guint input_id;

	if (opening) {
		GIOChannel *channel;
		/* Make sure we don't pass on these file descriptors to any
		   exec'ed children */
		fcntl (IceConnectionNumber (conn), F_SETFD,
		       fcntl (IceConnectionNumber (conn), F_GETFD, 0) | FD_CLOEXEC);
			
		channel = g_io_channel_unix_new (IceConnectionNumber (conn));
		input_id = g_io_add_watch (channel,
					   G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_PRI,
					   process_ice_messages,
					   conn);
		g_io_channel_unref (channel);

		*watch_data = (IcePointer) GUINT_TO_POINTER (input_id);
	} else {
		input_id = GPOINTER_TO_UINT ((gpointer) *watch_data);

		g_source_remove (input_id);
	}
}

static void
ice_io_error_handler (IceConn conn)
{
	if (ice_installed_handler)
		(*ice_installed_handler) (conn);
}

static void
ice_init (void)
{
	static gboolean initialized = FALSE;

	if (!initialized) {
		IceIOErrorHandler default_handler;

		ice_installed_handler = IceSetIOErrorHandler (NULL);
		default_handler = IceSetIOErrorHandler (ice_io_error_handler);

		if (ice_installed_handler == default_handler)
			ice_installed_handler = NULL;

		IceAddConnectionWatch (new_ice_connection, NULL);

		initialized = TRUE;
	}
}

static gchar *
tf_sm_get_session_manager_var (void)
{
	GDir   *dir;
	const gchar *filename;
	const gchar *dirname = "/tmp/.ICE-unix/";
	gchar *file = NULL;
	gchar *socket_uri;
	GError *error = NULL;

	dir = g_dir_open (dirname, 0, &error);
	if (!dir) {
		g_warning ("%s\n", error->message);
		g_error_free (error);

		return NULL;
	}

	while ((filename = g_dir_read_name (dir))) {
		struct stat file_info;

		file = g_build_filename (dirname, filename, NULL);

		if (g_stat (file, &file_info) < 0) {
			g_free (file);
			file = NULL;
			continue;
		}

		if (S_ISSOCK (file_info.st_mode) &&
		    file_info.st_uid == getuid () &&
		    g_ascii_isdigit (filename[0])) {
			break;
		}

		g_free (file);
		file = NULL;
	}

	g_dir_close (dir);

	if (!file)
		return NULL;

	socket_uri = g_strdup_printf ("local/%s:%s",
				      g_get_host_name (),
				      file);
	return socket_uri;
}

#define ERROR_LENGHT 256

/* Public methods */
void
tf_sm_init (TfSm *sm)
{
	SmcCallbacks  callbacks;
	gchar        *client_id;
	gchar        *tmp;
	gchar       **cmd;
	
	ice_init ();

	/* Ok, I also don't like it, if you know
	 * a better way to do it, please let me know
	 * and I will pay you a beer
	 */
	while (!(tmp = tf_sm_get_session_manager_var ()))
		sleep (1);
	
	g_setenv ("SESSION_MANAGER", tmp, FALSE);
	g_free (tmp);
	
	if (g_getenv ("SESSION_MANAGER")) {
		gchar error[ERROR_LENGHT] = "";
		
		callbacks.save_yourself.callback = save_yourself_callback;
		callbacks.save_complete.callback = save_complete_callback;
		callbacks.shutdown_cancelled.callback = shutdown_cancelled_callback;
		callbacks.die.callback = die_callback;
		
		callbacks.save_yourself.client_data = NULL;
		callbacks.die.client_data = NULL;
		callbacks.save_complete.client_data = NULL;
		callbacks.shutdown_cancelled.client_data = NULL;
		
		/* Connect to session manager */
		sm->priv->conn =
			SmcOpenConnection (NULL, sm,
					   SmProtoMajor, SmProtoMinor,
					   SmcSaveYourselfProcMask |
					   SmcDieProcMask          |
					   SmcSaveCompleteProcMask |
					   SmcShutdownCancelledProcMask,
					   &callbacks,
					   NULL, &client_id,
					   ERROR_LENGHT, error);
		
		if (error[0])
			g_warning ("%s\n", error);
	}

	g_free (client_id);

	if (!sm->priv->conn) {
		return;
	}

	/* Current directory */
	tmp = g_get_current_dir ();
	_session_set_string (sm->priv->conn, SmCurrentDirectory, tmp);
	g_free (tmp);

	/* PID */
	tmp = g_strdup_printf ("%d", (int) getpid ());
	_session_set_string (sm->priv->conn, SmProcessID, tmp);
	g_free(tmp);

	/* User name */
	tmp = g_strdup (g_get_user_name ());
	_session_set_string (sm->priv->conn, SmUserID, tmp);
	g_free (tmp);

	/* The daemon is already started by xsession */
	_session_set_char (sm->priv->conn, SmRestartStyleHint, (gchar) SmRestartNever);

	/* Program Name */
	_session_set_string (sm->priv->conn, SmProgram, TF_APP_NAME);

	/* Clone Command */
	cmd = g_new (gchar *, 2);
	cmd[0] = g_strdup (sm->priv->command);
	cmd[1] = NULL;
	_session_set_array (sm->priv->conn, SmCloneCommand, cmd);

	/* Restart Command */
	_session_set_array (sm->priv->conn, SmRestartCommand, cmd);
	g_strfreev (cmd);
	
	/* Discard Command */
	cmd = g_new (gchar *, 2);
	cmd[0] = g_strdup ("/bin/true");
	cmd[1] = NULL;
	_session_set_array (sm->priv->conn, SmDiscardCommand, cmd);
	g_strfreev (cmd);
}
