/*
 *  Copyright (C) 2010 Edmundo Álvarez Jiménez
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Authors: Edmundo Álvarez Jiménez <e.alvarezj@gmail.com>
 */

#include <glib.h>
#include <glib/gstdio.h>
#include <unistd.h>

#include "tf-xml-validator.h"
#include "tf-application.h"

enum {
    PROP_0,
    PROP_XML
};

struct _TfXmlValidatorPrivate {
    GIOChannel      *xml;
};


#define TF_XML_VALIDATOR_GET_PRIVATE(object) \
        (G_TYPE_INSTANCE_GET_PRIVATE((object), TF_TYPE_XML_VALIDATOR, TfXmlValidatorPrivate))

#define XML_DOC_DEFINITION "<?xml version=\"1.0\"?>\n"
#define XML_ROOT_OPENING_TAG "<events>\n"
#define XML_ROOT_CLOSING_TAG "</events>\n"

static GObjectClass *parent_class = NULL;

static void tf_xml_validator_init         (TfXmlValidator      *val);
static void tf_xml_validator_class_init   (TfXmlValidatorClass *klass);
static void tf_xml_validator_finalize     (GObject             *object);
static void tf_xml_validator_set_property (GObject             *object,
                guint        prop_id,
                const GValue *value,
                GParamSpec   *pspec);


GType
tf_xml_validator_get_type (void)
{
    static GType type = 0;

    if (!type) {
        static const GTypeInfo info = {
            sizeof (TfXmlValidatorClass),
            (GBaseInitFunc) NULL,
            (GBaseFinalizeFunc) NULL,
            (GClassInitFunc) tf_xml_validator_class_init,
            NULL,
            NULL,
            sizeof (TfXmlValidator),
            0,
            (GInstanceInitFunc) tf_xml_validator_init
        };

        type = g_type_register_static (G_TYPE_OBJECT, "TfXmlValidator",
                &info, 0);
    }

    return type;
}


static void
tf_xml_validator_set_property (GObject *object,
        guint           prop_id,
        const GValue    *value,
        GParamSpec      *pspec)
{
    TfXmlValidator *val = TF_XML_VALIDATOR (object);

    switch (prop_id) {
    case PROP_XML:
        g_free (val->priv->xml);
        val->priv->xml = g_value_get_pointer (value);
        g_assert (val->priv->xml != NULL);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}


static void
tf_xml_validator_init (TfXmlValidator *val)
{
    val->priv = TF_XML_VALIDATOR_GET_PRIVATE (val);

    val->priv->xml = NULL;
}


static void
tf_xml_validator_class_init (TfXmlValidatorClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->set_property = tf_xml_validator_set_property;

    parent_class = g_type_class_peek_parent (klass);

    g_type_class_add_private (klass, sizeof (TfXmlValidatorPrivate));

    g_object_class_install_property (object_class,
            PROP_XML,
            g_param_spec_pointer ("xml",
                    "XML Channel",
                    "XML channel to check",
                    G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY));

    object_class->finalize = tf_xml_validator_finalize;
}


static void
tf_xml_validator_finalize (GObject *object)
{
    TfXmlValidator *validator = TF_XML_VALIDATOR (object);

    if (validator->priv->xml) {
        g_io_channel_unref (validator->priv->xml);
        validator->priv->xml = NULL;
    }

    if (G_OBJECT_CLASS (parent_class)->finalize)
        (* G_OBJECT_CLASS (parent_class)->finalize) (object);
}


TfXmlValidator *
tf_xml_validator_new (const gchar *filename)
{
    GIOChannel      *file;
    GError          *error = NULL;
    TfXmlValidator  *val;

    file = g_io_channel_new_file (filename, "r", &error);
    if (error) {
        g_critical ("%s", error->message);
        g_error_free (error);
        return NULL;
    }

    val = TF_XML_VALIDATOR (g_object_new (TF_TYPE_XML_VALIDATOR,
            "xml", file, NULL));

    return val;
}


/* Validates an XML file */
/* TODO: Validate XML against a XML Schema or DTD */
gboolean
validate (TfXmlValidator *val)
{
    gchar     *line;
    GIOStatus status;
    GError    *error = NULL;
    gint      i = 0;
    gboolean  valid;

    debug ("Validating XML...\n");

    for (i = 0; i < 3; i++) {
        status = g_io_channel_read_line (val->priv->xml, &line, NULL, NULL,
                &error);

        if (status != G_IO_STATUS_NORMAL) {
            g_warning ("read file failed with status %d\n", status);
            g_error_free (error);
            return FALSE;
        }

        debug ("line: %s\n", line);

        switch (i) {
        case 0:
            valid = (g_strcmp0(line, XML_DOC_DEFINITION) == 0);
            break;
        case 1:
            valid = (g_strcmp0(line, XML_ROOT_OPENING_TAG) == 0);

            g_io_channel_seek_position (val->priv->xml,
                    g_utf8_strlen (XML_ROOT_CLOSING_TAG, -1) * -1,
                    G_SEEK_END, &error);

            if (error) {
                g_error_free (error);
                g_free (line);
                return FALSE;
            }
                
            break;
        case 2:
            valid = (g_strcmp0(line, XML_ROOT_CLOSING_TAG) == 0);

            g_io_channel_seek_position (val->priv->xml, 0, G_SEEK_SET, &error);

            if (error)
                g_error_free (error);

            break;
        }

        g_free (line);

        if (!valid) {
            debug ("Malformed doc\n");
            return FALSE;
        }
    }

    debug ("Well-formed doc\n");

    return TRUE;
}
