/*
 *  Copyright (C) 2006 Carlos Garcia Campos
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Authors : Carlos Garcia Campos <carlosgc@gsyc.escet.urjc.es>
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <glib/gprintf.h>

#include "tf-xutils.h"
#include "tf-application.h"

typedef struct _TfXErrorTrap TfXErrorTrap;

struct _TfXErrorTrap {
	gint (*old_handler) (Display *, XErrorEvent *);
	gint error_warnings;
	gint error_code;
};

static GSList *x_error_traps = NULL;               /* List of error traps */
static GSList *x_error_trap_free_list = NULL;      /* Free list */
static gint    x_error_code = 0;
static gint    x_error_warnings = TRUE;

static gint
tf_x_error (Display *display, XErrorEvent *error)
{
	if (error->error_code) {
		if (x_error_warnings) {
			gchar buf[64];
			gchar *msg;

			XGetErrorText (display, error->error_code, buf, 63);

			msg = g_strdup_printf ("The program '%s' received an X Window System error.\n"
					       "This probably reflects a bug in the program.\n"
					       "The error was '%s'.\n"
					       "  (Details: serial %ld error_code %d request_code %d minor_code %d)",
					       g_get_prgname (), buf,
					       error->serial,
					       error->error_code,
					       error->request_code,
					       error->minor_code);
			
			g_fprintf (stderr, "%s\n", msg);
			g_free (msg);
			
			tf_application_quit ();
			exit (1);
		}
		x_error_code = error->error_code;
	}
	
	return 0;
}

static gint
tf_x_io_error (Display *display)
{
	if (errno == EPIPE) {
		g_fprintf (stderr,
			   "The application '%s' lost its connection to the display %s;\n"
			   "most likely the X server was shut down or you killed/destroyed\n"
			   "the application.\n",
			   g_get_prgname (), DisplayString (display));
	} else {
		g_fprintf (stderr, "%s: Fatal IO error %d (%s) on X server %s.\n",
			   g_get_prgname (),
			   errno, g_strerror (errno),
			   DisplayString (display));
	}

	tf_application_quit ();
	exit (1);
}

void
tf_x_error_init (void)
{
	XSetErrorHandler (tf_x_error);
	XSetIOErrorHandler (tf_x_io_error);
}

static Atom
_atom_get (Display *display, const char *atom_name)
{
	g_assert (atom_name != NULL);
	
	return XInternAtom (display, atom_name, FALSE);
}

Atom
_active_window_atom (Display *display)
{
	static Atom active_window = None;

	if (active_window == None) {
		active_window = _atom_get (display, "_NET_ACTIVE_WINDOW");
	}

	return active_window;
}

static Atom
_utf8_atom (Display *display)
{
	static Atom utf8_atom = None;

	if (utf8_atom == None) {
		utf8_atom = _atom_get (display, "UTF8_STRING");
	}

	return utf8_atom;
}

static Atom
_string_atom (Display *display)
{
	static Atom string_atom = None;

	if (string_atom == None) {
		string_atom = _atom_get (display, "STRING");
	}

	return string_atom;
}

static Atom
_visible_name_atom (Display *display)
{
	static Atom visible_name_atom = None;

	if (visible_name_atom == None) {
		visible_name_atom = _atom_get (display, "_NET_WM_VISIBLE_NAME");
	}

	return visible_name_atom;
}

static Atom
_name_atom (Display *display)
{
	static Atom name_atom = None;

	if (name_atom == None) {
		name_atom = _atom_get (display, "_NET_WM_NAME");
	}

	return name_atom;
}

/* Taken from GDK */
static void
x_error_trap_push (void)
{
	GSList       *node;
	TfXErrorTrap *trap;
	
	if (x_error_trap_free_list) {
		node = x_error_trap_free_list;
		x_error_trap_free_list = x_error_trap_free_list->next;
	} else {
		node = g_slist_alloc ();
		node->data = g_new (TfXErrorTrap, 1);
	}

	node->next = x_error_traps;
	x_error_traps = node;

	trap = node->data;
	trap->old_handler = XSetErrorHandler (tf_x_error);
	trap->error_code = x_error_code;
	trap->error_warnings = x_error_warnings;

	x_error_code = 0;
	x_error_warnings = 0;
}

/* Taken from GDK */
static gint
x_error_trap_pop (Display *display)
{
	GSList       *node;
	TfXErrorTrap *trap;
	gint          result;

	g_return_val_if_fail (x_error_traps != NULL, 0);

	XSync (display, False);

	node = x_error_traps;
	x_error_traps = x_error_traps->next;

	node->next = x_error_trap_free_list;
	x_error_trap_free_list = node;

	result = x_error_code;

	trap = node->data;
	x_error_code = trap->error_code;
	x_error_warnings = trap->error_warnings;
	XSetErrorHandler (trap->old_handler);

	return result;
}

static gchar *
get_utf8_property (Display *display, Window win, Atom atom)
{
	Atom    type;
	gint    format;
	gulong  nitems;
	gulong  bytes_after;
	guchar *val;
	gint    result, err;
	char   *retval;
	Atom    utf8_string;

	utf8_string = _utf8_atom (display);

	x_error_trap_push ();
	
	type = None;
	val = NULL;
	result = XGetWindowProperty (display,
				     win,
				     atom,
				     0, G_MAXLONG,
				     False, utf8_string,
				     &type, &format, &nitems,
				     &bytes_after, (guchar **)&val);
	
	err = x_error_trap_pop (display);

	if (err != Success || result != Success)
		return NULL;

	if (type != utf8_string ||
	    format != 8 || nitems == 0) {
		if (val)
			XFree (val);
		return NULL;
	}

	if (!g_utf8_validate ((gchar *)val, nitems, NULL)) {
		g_warning ("Property contained invalid UTF-8\n");
		XFree (val);
		return NULL;
	}

	retval = g_strndup ((gchar *)val, nitems);

	XFree (val);

	return retval;
}

static gchar *
text_property_to_utf8 (Display *display, const XTextProperty *text)
{
	gchar  *str = NULL;
	GError *error = NULL;
	
	if (text->encoding == _string_atom (display)) {
		str = g_convert ((gchar *)text->value, text->nitems,
				 "utf-8", "iso-8859-1",
				 NULL, NULL, &error);
	} else if (text->encoding == _utf8_atom (display)) {
		str = g_strndup ((gchar *)text->value, text->nitems);
		if (!g_utf8_validate (str, -1, NULL)) {
			g_warning ("Error converting selection from utf-8 string");
			g_free (str);
			str = NULL;
		}
	} else {
		str = g_locale_to_utf8 ((gchar *)text->value,
					text->nitems, NULL, NULL,
					&error);
	}

	if (error) {
		g_warning ("%s", error->message);
		g_error_free (error);
	}

	return str;
}

static gchar *
get_text_property (Display *display, Window win, Atom atom)
{
	XTextProperty text;
	gchar        *retval = NULL;

	x_error_trap_push ();

	text.nitems = 0;
	if (XGetTextProperty (display, win, &text, atom)) {
		if (text.value) {
			retval = text_property_to_utf8 (display, &text);
			XFree (text.value);
		}
	} 

	x_error_trap_pop (display);

	return retval;
}

Window
tf_get_active_window (Display *display)
{
	Atom    type, active_window;
	gint    format;
	gulong  nitems;
	gulong  bytes_after;
	Window *w;
	gint    result, err;
	Window  retval;

	x_error_trap_push ();
	
	type = None;
	active_window = _active_window_atom (display);
	
	result = XGetWindowProperty (display,
				     DefaultRootWindow (display),
				     active_window,
				     0, G_MAXLONG,
				     False, XA_WINDOW, &type, &format, &nitems,
				     &bytes_after, (guchar **)&w);
	
	err = x_error_trap_pop (display);
	if (err != Success || result != Success)
		return None;

	if (type != XA_WINDOW) {
		XFree (w);
		return None;
	}

	retval = *w;

	XFree (w);

	return retval;
}

gchar *
tf_get_application_name (Display *display, Window win)
{
	XClassHint classhint;
	gchar     *app_name = NULL;

	x_error_trap_push ();

	classhint.res_name = NULL;
	classhint.res_class = NULL;

	XGetClassHint (display, win, &classhint);

	x_error_trap_pop (display);
	
	if (classhint.res_class) {
		app_name = g_strdup (classhint.res_class);
		XFree (classhint.res_class);
	}
	
	if (classhint.res_name) {
		if (!app_name) {
			app_name = g_strdup (classhint.res_name);
		}
		XFree (classhint.res_name);
	}

	return app_name;
}

gchar *
tf_get_window_title (Display *display, Window win)
{
	gchar *name;

	name = get_utf8_property (display, win, _visible_name_atom (display));

	if (name)
		return name;

	name = get_utf8_property (display, win, _name_atom (display));

	if (name)
		return name;
	
	name = get_text_property (display, win, XA_WM_NAME);

	return name;
}

/* Session management helper functions
 * (Taken from gnome-client)
 */
static void
session_set_value (SmcConn      conn,
		   const gchar *name,
		   const char  *type,
		   int          num_vals,
		   SmPropValue *vals)
{
	SmProp *proplist[1];
	SmProp prop;

	g_assert (conn != NULL);

	prop.name = (gchar *)name;
	prop.type = (gchar *)type;
	prop.num_vals = num_vals;
	prop.vals = vals;

	proplist[0] = &prop;

	SmcSetProperties (conn, 1, proplist);
}

void
_session_set_string (SmcConn      conn,
		     const gchar *name,
		     const gchar *value)
{
	SmPropValue val;

	g_assert (name != NULL);

	val.length = strlen (value) + 1;
	val.value  = (gchar *)value;

	session_set_value (conn, name, SmARRAY8, 1, &val);
}

void
_session_set_char (SmcConn      conn,
		   const gchar *name,
		   gchar        value)
{
	SmPropValue val;

	g_assert (name != NULL);

	val.length = 1;
	val.value  = &value;

	session_set_value (conn, name, SmCARD8, 1, &val);
}

void
_session_set_array (SmcConn      conn,
		    const gchar *name,
		    gchar      **array)
{
	gint    argc;
	gchar **ptr;
	gint    i;

	SmPropValue *vals;

	g_assert (name != NULL);

	/* We count the number of elements in our array.  */
	for (ptr = array, argc = 0; *ptr ; ptr++, argc++) /* LOOP */;

	/* Now initialize the 'vals' array.  */
	vals = g_new (SmPropValue, argc);
	for (ptr = array, i = 0 ; i < argc ; ptr++, i++) {
		vals[i].length = strlen (*ptr);
		vals[i].value  = *ptr;
	}

	session_set_value (conn, name, SmLISTofARRAY8, argc, vals);

	g_free (vals);
}
