/*
 *  Copyright (C) 2006 Carlos Garcia Campos
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Authors : Carlos Garcia Campos <carlosgc@gsyc.escet.urjc.es>
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <errno.h>

typedef enum {
	FALSE,
	TRUE
} boolean;

static boolean verbose_mode = FALSE;

static void
usage (void) 
{
	fprintf (stderr, "tf-launch [--version] [--help] [--verbose] \n");
	exit (0);
}

static void
version (void)
{
	printf ("DeTraS TempusFugit Launcher %s\n"
		"Copyright (C) 2006 Carlos Garcia Campos\n"
		"This is free software; see the source for copying conditions.\n"
		"There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n",
		VERSION);
	exit (0);
}

static void
verbose (const char *format,
	 ...)
{
	va_list args;
	static boolean verbose_initted = FALSE;
	static boolean verbose = TRUE;

	if (!verbose)
		return;

	if (!verbose_initted) {
		verbose = verbose_mode || getenv ("TEMPUSFUGIT_VERBOSE");

		verbose_initted = TRUE;

		if (!verbose)
			return;
	}

	fprintf (stderr, "%lu: ", (unsigned long) getpid ());

	va_start (args, format);
	vfprintf (stderr, format, args);
	va_end (args);
}

int main (int argc, char **argv)
{
	int         remaining_args = 0;
	const char *runprog = NULL;
	int         i;
	pid_t       pid;

	/* Process command line arguments */
	i = 1;
	while (i < argc) {
		const char *arg = argv[i];

		if (strcmp (arg, "--help") == 0 ||
		    strcmp (arg, "-h") == 0 ||
		    strcmp (arg, "-?") == 0) {
			usage ();
		} else if (strcmp (arg, "--version") == 0) {
			version ();
		} else if (strcmp (arg, "--verbose") == 0 ||
			   strcmp (arg, "-v") == 0) {
			verbose_mode = TRUE;
		} else {
			runprog = arg;
			remaining_args = i + 1;

			break;
		}

		i++;
	}

	switch ((pid = fork ())) {
	case -1:
		/* Error */
		fprintf (stderr, "Error creating new process\n");
		break;
	case 0:
		/* Child */
		switch (fork ()) {
		case -1:
			/* Error */
			fprintf (stderr, "Error creating new process\n");
			break;
		case 0:
			/* Child: exec TempusFugit */
			verbose ("exec TempusFugit\n");
			if (verbose_mode) {
				execl (TF_PATH, "tempusfugit", "-v", NULL);
			} else {
				execl (TF_PATH, "tempusfugit", NULL);
			}

			exit (1);
			       
			break;
		default:
			/* Parent: I die and init will adopt my child */
			verbose ("I die and init will adopt my child\n");
			exit (0);
		}

		break;
	default: {
		/* Parent: exec commands */
		int ret;

		verbose ("Waiting for my child's death\n");
		do {
			ret = waitpid (pid, NULL, 0);
		} while (ret < 0 && errno == EINTR);
		
		if (runprog) {
			char **args;

			args = (char **) malloc (sizeof (char *) * ((argc - remaining_args) + 2));

			args[0] = strdup (runprog);
			for (i = 1; i <= (argc - remaining_args); i++) {
				int j = remaining_args + i - 1;
				size_t len = strlen (argv[j]) + 1;
				
				args[i] = (char *) malloc (sizeof (char) * (len));
				strncpy (args[i], argv[j], len);
			}
			args[i] = NULL;

			verbose ("exec remainig commands\n");
			execvp (runprog, args);

			exit (1);
		}
	}

		exit (0);
	}

	return 0;
}
